#include <gba.h>
#include <bg.h>
#include <dmachain.h>
#include <pcgcache.h>
#include <sprite.h>
#include <gamemath.h>
#include "bioscall.h"
#include "disp.h"
#include "pdrv.h"
#include "nmid.h"
#include "res/dsp.h"
#include "res/dsp_spr.h"
#include "res/ascii.h"
#include "res/_MapTable.h"

#define DIGIT_CHIP		0x30
#define HEX_CHIP		0x8c
#define ALPHABET_CHIP	0x41
#define ASCII_CHIP		0x20
#define TRACK_CHIP		0x80
#define MONO_CHIP		214
#define POLY_CHIP		215
#define UPDOWN_CHIP		273
#define VOL_CHIP		276
#define KEY_CHIP		278
#define BPM_CHIP		280
#define PLAY_CHIP		8
#define STOP_CHIP		11
#define BACK_CHIP		23

#define SPR_PRI_KEYBOARD	0x80

vu16 *objpal= OBJpal;
vu16 *bgpal = MEM_BG_PAL;		// BGpbg AhX
vu16 *tiles = MEM_BG_CHAR(0);	// ^CLN^f[^ AhX
vu16 *map0  = MEM_BG_MAP(28);	// ^C}bvf[^ AhX
vu16 *map1  = MEM_BG_MAP(29);	// ^C}bvf[^ AhX
vu16 *map2  = MEM_BG_MAP(30);	// ^C}bvf[^ AhX
vu16 *map3  = MEM_BG_MAP(31);	// ^C}bvf[^ AhX

SPRWORK *key_spr[MAX_CHANNELS];

extern PDRVWORK pdrvw;
extern INSTBANK inst[];

static const struct {
	u8 x, y;
	u16 num;
	u16 chip[4];
} CursorTable[] = {
//	  x   y  n
	{ 0,  2, 3, {TRACK_CHIP, TRACK_CHIP+1, HEX_CHIP + 0, } }, // trk0
	{ 0,  3, 3, {TRACK_CHIP, TRACK_CHIP+1, HEX_CHIP + 1, } }, // trk1
	{ 0,  4, 3, {TRACK_CHIP, TRACK_CHIP+1, HEX_CHIP + 2, } }, // trk2
	{ 0,  5, 3, {TRACK_CHIP, TRACK_CHIP+1, HEX_CHIP + 3, } }, // trk3
	{ 0,  6, 3, {TRACK_CHIP, TRACK_CHIP+1, HEX_CHIP + 4, } }, // trk4
	{ 0,  7, 3, {TRACK_CHIP, TRACK_CHIP+1, HEX_CHIP + 5, } }, // trk5
	{ 0,  8, 3, {TRACK_CHIP, TRACK_CHIP+1, HEX_CHIP + 6, } }, // trk6
	{ 0,  9, 3, {TRACK_CHIP, TRACK_CHIP+1, HEX_CHIP + 7, } }, // trk7
	{ 0, 10, 3, {TRACK_CHIP, TRACK_CHIP+1, HEX_CHIP + 8, } }, // trk8
	{ 0, 11, 3, {TRACK_CHIP, TRACK_CHIP+1, HEX_CHIP + 9, } }, // trk9
	{ 0, 12, 3, {TRACK_CHIP, TRACK_CHIP+1, HEX_CHIP + 10, } }, // trkA
	{ 0, 13, 3, {TRACK_CHIP, TRACK_CHIP+1, HEX_CHIP + 11, } }, // trkB
	{ 0, 14, 3, {TRACK_CHIP, TRACK_CHIP+1, HEX_CHIP + 12, } }, // trkC
	{ 0, 15, 3, {TRACK_CHIP, TRACK_CHIP+1, HEX_CHIP + 13, } }, // trkD
	{ 0, 16, 3, {TRACK_CHIP, TRACK_CHIP+1, HEX_CHIP + 14, } }, // trkE
	{ 0, 17, 3, {TRACK_CHIP, TRACK_CHIP+1, HEX_CHIP + 15, } }, // trkF
	{ 0, 18, 2, {VOL_CHIP, VOL_CHIP+1, } }, // Vol
	{ 4, 18, 2, {KEY_CHIP, KEY_CHIP+1, } }, // Key
	{ 8, 18, 2, {BPM_CHIP, BPM_CHIP+1, } }, // BPM
	{ 13, 18, 3, {BACK_CHIP, BACK_CHIP+1, BACK_CHIP+2, } }, // Back
	{ 16, 18, 3, {PLAY_CHIP, PLAY_CHIP+1, PLAY_CHIP+2, } }, // Play
	{ 19, 18, 3, {STOP_CHIP, STOP_CHIP+1, STOP_CHIP+2, } }, // Stop
};

static const struct {
	u8 gauge[2];
} GaugeTable[] = {
	{{0xb3,0xb3}}, // 0
	{{0xb2,0xb3}}, // 1
	{{0xb1,0xb3}}, // 2
	{{0xb0,0xb3}}, // 3
	{{0xaf,0xb3}}, // 4
	{{0xaf,0xb2}}, // 5
	{{0xaf,0xb1}}, // 6
	{{0xaf,0xb0}}, // 7
	{{0xaf,0xaf}}, // 8
	{{0xaf,0xaf}}, // 9
	{{0xaf,0xaf}}, // 10
	{{0xaf,0xaf}}, // 11
	{{0xaf,0xaf}}, // 12
	{{0xaf,0xaf}}, // 13
	{{0xaf,0xaf}}, // 14
	{{0xaf,0xaf}}, // 15
	{{0xaf,0xaf}}, // 16
};

static const u8 GaugeTable2[] = {
	0xf6, // 0
	0xf7, // 1
	0xf8, // 2
	0xf8, // 3
	0xf9, // 4
	0xf9, // 5
	0xfa, // 6
	0xfa, // 7
	0xfb, // 8
	0xfb, // 9
	0xfc, // 10
	0xfc, // 11
	0xfd, // 12
	0xfd, // 13
	0xfe, // 14
	0xfe, // 15
	0xff, // 16
};


static const struct {
	u8 x;
	u8 pat;
} KeyTable[] = {
//      C      C#       D      D#       E       F      F#       G      G#       A       A#        B
	{ 9,0}, { 9,1}, { 9,2}, { 9,3}, { 9,4}, { 9,5}, { 9,6}, { 9,7}, { 9,8}, { 9,9}, { 9,10}, { 9,11}, // o-2    0 -  11
	{10,0}, {10,1}, {10,2}, {10,3}, {10,4}, {10,5}, {10,6}, {10,7}, {10,8}, {10,9}, {10,10}, {10,11}, // o-1   12 -  23
	{ 1,0}, { 1,1}, { 1,2}, { 1,3}, { 1,4}, { 1,5}, { 1,6}, { 1,7}, { 1,8}, { 1,9}, { 1,10}, { 1,11}, // o0    24 -  35
	{ 2,0}, { 2,1}, { 2,2}, { 2,3}, { 2,4}, { 2,5}, { 2,6}, { 2,7}, { 2,8}, { 2,9}, { 2,10}, { 2,11}, // o1    36 -  47
	{ 3,0}, { 3,1}, { 3,2}, { 3,3}, { 3,4}, { 3,5}, { 3,6}, { 3,7}, { 3,8}, { 3,9}, { 3,10}, { 3,11}, // o2    48 -  59
	{ 4,0}, { 4,1}, { 4,2}, { 4,3}, { 4,4}, { 4,5}, { 4,6}, { 4,7}, { 4,8}, { 4,9}, { 4,10}, { 4,11}, // o3    60 -  71
	{ 5,0}, { 5,1}, { 5,2}, { 5,3}, { 5,4}, { 5,5}, { 5,6}, { 5,7}, { 5,8}, { 5,9}, { 5,10}, { 5,11}, // o4    72 -  13
	{ 6,0}, { 6,1}, { 6,2}, { 6,3}, { 6,4}, { 6,5}, { 6,6}, { 6,7}, { 6,8}, { 6,9}, { 6,10}, { 6,11}, // o5    84 -  95
	{ 7,0}, { 7,1}, { 7,2}, { 7,3}, { 7,4}, { 7,5}, { 7,6}, { 7,7}, { 7,8}, { 7,9}, { 7,10}, { 7,11}, // o6    96 - 107
	{ 8,0}, { 8,1}, { 8,2}, { 8,3}, { 8,4}, { 8,5}, { 8,6}, { 8,7}, { 8,8}, { 8,9}, { 8,10}, { 8,11}, // o7   108 - 119
	{ 9,0}, { 9,1}, { 9,2}, { 9,3}, { 9,4}, { 9,5}, { 9,6}, { 9,7}, { 9,8}, { 9,9}, { 9,10}, { 9,11}, // o8   120 - 127
};


//--------------------------------------------------------------
//
//--------------------------------------------------------------
void 
DispInit()
{
	u32 i;
	u32 num;
	REG_DISPCNT = DISP_MODE_0 | DISP_BG0_ON | DISP_BG1_ON | DISP_OBJ_ON | DISP_OBJ_1D_MAP;
	REG_BG0CNT = ( BG_SIZEA_256_256 | BG_COLOR_16 | BG_CHARBASE(0) | BG_MAPBASE(28) | 0);
	REG_BG1CNT = ( BG_SIZEA_256_256 | BG_COLOR_16 | BG_CHARBASE(0) | BG_MAPBASE(29) | 1);
	REG_BG2CNT = ( BG_SIZEA_256_256 | BG_COLOR_16 | BG_CHARBASE(0) | BG_MAPBASE(30) | 2 | BG_OVERLAP);
	REG_BG3CNT = ( BG_SIZEA_256_256 | BG_COLOR_16 | BG_CHARBASE(0) | BG_MAPBASE(31) | 3 | BG_OVERLAP);

	// BG Palette
	for(i = 0; i < 16*4; i++) 
		bgpal[i] = dsp_pal[i];
	// BG Tile
	for(i = 0; i < 2*8*512; i++) 
		tiles[i] = dsp_chr[i];
	// Map
	for(i=0;i<32*20;i++) {
		map0[i] = map0_pattern[i];
		map1[i] = map1_pattern[i];
	}
	// OBJ Palette
	DMAChain_Enqueue((u32*)&dsp_spr_pal[0], (u32*)OBJpal, 16*16);

	// Sprites
	SpriteSetView(0, 0);
	for (i = 0; i < MAX_CHANNELS; i++) {
		key_spr[i] = SpriteSet(SPR_PRI_KEYBOARD);
		SpritePattern(key_spr[i], 0);
		SpriteChangePallete( key_spr[i], 0 );
		SpriteChangePriority( key_spr[i], 1 );
		key_spr[i]->disp = 1;
	}
	PrintHex(map0, pdrvw.master_vol, 2, 2, 18, 0);
	PrintHex(map0, pdrvw.master_keyshift & 0xff, 2, 6, 18, 0);
	PrintDigit0I(map0, nmidi.bpm, 3, 10, 18, 0);

	for (num = CUR_BACK; num < CUR_NONE; num++) {
		for (i = 0; i < CursorTable[num].num; i++) {
			map0[CursorTable[num].x + i + 32 * CursorTable[num].y] = CursorTable[num].chip[i];
		}
	}

}

//--------------------------------------------------------------
void 
TrackDispUpdate()
{
	u16 objnum = 0;
	u16 i,j;

	for (i = 0; i < NMID_MAX_CHANNELS; i++) {
		NmidTrackStatus *trk = &nmidi.status[i];
		u8 mod = nmidi.status[i].lfo_depth;

		// Disp Modulation
		if (mod != 0 && mod < 16) {
			mod = 16;
		}
		mod /= 16;
		map0[(i + 2)*32 + 29] = GaugeTable2[mod];

		// Disp Mono/Poly
		if (trk->bit.monopoly)
			map0[(i + 2)*32 + 28] = MONO_CHIP;
		else
			map0[(i + 2)*32 + 28] = POLY_CHIP;

		// Set Keyboard (Mute or Unmute)
		if (trk->bit.mute) {
			for (j = 0; j < 32; j++) {
				map1[(i + 2)*32 + j] = map1_mute[j];
			}
		} else {
			for (j = 0; j < 32; j++) {
				map1[(i + 2)*32 + j] = map1_play[j];
			}
		}

		// Disp Keyboard
		for (j = 0; j < trk->voicenum; j++) {
			u8 note = trk->voice[j].note;
			s16 pitch = trk->voice[j].porta_detune;
			pitch += (note << 8);
			note = (pitch >> 8);
			if (trk->bit.mute) {
				SpriteChangePallete(key_spr[objnum], 1);
			} else {
				SpriteChangePallete(key_spr[objnum], 0);
			}
			SpritePattern(key_spr[objnum], KeyTable[note].pat);
			SpriteMove(key_spr[objnum], KeyTable[note].x * 24, i * 8 + 16);
			key_spr[objnum]->disp = 1;
			objnum++;
		}
	}
	for (;objnum < MAX_CHANNELS; objnum++) {
		key_spr[objnum]->disp = 0;
	}
}

void
PrintTime(u32 count)
{
	u8 digit = 0;
	s32 msec_h, msec_l, sec_h, sec_l, min_h, min_l;
	s32 msec, sec, min;

	sec = bios_divmod(count, 60, (int*)&msec);
	min = bios_divmod(sec, 60, (int*)&sec);

	msec_h = bios_divmod(msec, 10, (int*)&msec_l);
	sec_h = bios_divmod(sec, 10, (int*)&sec_l);
	min_h = bios_divmod(min, 10, (int*)&min_l);

	map0[32*0 + 15+digit] = map0_num[min_h][0];
	map0[32*1 + 15+digit] = map0_num[min_h][1]; digit++;
	map0[32*0 + 15+digit] = map0_num[min_l][0];
	map0[32*1 + 15+digit] = map0_num[min_l][1]; digit++;
	digit++;
	map0[32*0 + 15+digit] = map0_num[sec_h][0];
	map0[32*1 + 15+digit] = map0_num[sec_h][1]; digit++;
	map0[32*0 + 15+digit] = map0_num[sec_l][0];
	map0[32*1 + 15+digit] = map0_num[sec_l][1]; digit++;
	digit++;
	map0[32*0 + 15+digit] = map0_num[msec_h][0];
	map0[32*1 + 15+digit] = map0_num[msec_h][1]; digit++;
	map0[32*0 + 15+digit] = map0_num[msec_l][0];
	map0[32*1 + 15+digit] = map0_num[msec_l][1]; digit++;
}

void 
PrintCpu(u32 num)
{
	if (num == 100) {
		map0[32*1 + 27] = DIGIT_CHIP + 1 + 4096;
		map0[32*1 + 28] = DIGIT_CHIP + 0 + 4096;
		map0[32*1 + 29] = DIGIT_CHIP + 0 + 4096;
	} else if (num >= 10) {
		s32 n;
		n = bios_divmod(num, 10, &num);
		map0[32*1 + 27] = DIGIT_CHIP + 4096;
		map0[32*1 + 28] = DIGIT_CHIP + 4096 + n;
		map0[32*1 + 29] = DIGIT_CHIP + 4096 + num;
	} else {
		map0[32*1 + 27] = DIGIT_CHIP + 4096;
		map0[32*1 + 28] = DIGIT_CHIP + 4096;
		map0[32*1 + 29] = DIGIT_CHIP + 4096 + num;
	}
}

void 
PrintRate(u32 num)
{
	if (num >= 10)
		num = 9;
	map0[26] = map0_rate[num][0] + 4096;
	map0[27] = map0_rate[num][1] + 4096;
	map0[28] = map0_rate[num][2] + 4096;
	map0[29] = map0_rate[num][3] + 4096;
}

//--------------------------------------------------------------
void 
PrintCursor(u8 num, u8 pal)
{
	int chip = 0;
	int i;
	u8 *c;
	
	num = num - CUR_TRK_0;
	for (i = 0; i < CursorTable[num].num; i++) {
		map0[CursorTable[num].x + i + 32 * CursorTable[num].y] = CursorTable[num].chip[i] + (pal * 4096);
	}

	switch (num) {
	case CUR_TRK_0 ... CUR_TRK_F:
		break;
	case CUR_VOL:
		PrintHex(map0, pdrvw.master_vol, 2, 2, 18, pal);
		break;
	case CUR_KEY:
		PrintHex(map0, pdrvw.master_keyshift & 0xff, 2, 6, 18, pal);
		break;
	case CUR_BPM:
		PrintDigit0I(map0, nmidi.bpm, 3, 10, 18, pal);
		break;
	default:
		break;
	}
}

//--------------------------------------------------------------
void 
PrintStr(u16 *map, const char* chr, u8 x, u8 y, u8 pal)
{
	u32 p = x + y * 32;
	while (*chr != 0x00) {
		u8 s = *chr - 0x20 + ASCII_CHIP;
		map[p] = s + pal * 8192;
		chr++;
		p++;
	}
}
void 
PrintDigit0I(u16 *map, const s32 num, u8 digit, s16 x, s16 y, u8 pal)
{
	s32 stemp = num;
	if (num < 0) {
		stemp *= -1;
		map[x + y*32 - 1] = 13 + pal * 8192;
	}
	do {
		map[x + y*32 + (digit-1)] = (stemp % 10) + HEX_CHIP + pal * 4096;
		if (digit == 0) break;
		digit--;
		stemp /= 10;
	} while (stemp > 0);

	while (digit > 0) {
		map[x + y*32 + (digit-1)] = HEX_CHIP + pal * 4096;
		digit--;
	}
	return;
}
void 
PrintDigit0(u16 *map, const s32 num, u8 digit, s16 x, s16 y, u8 pal)
{
	s32 stemp = num;
	if (num < 0) {
		stemp *= -1;
		map[x + y*32 - 1] = 13 + pal * 8192;
	}
	do {
		map[x + y*32 + (digit-1)] = (stemp % 10) + DIGIT_CHIP + pal * 4096;
		if (digit == 0) break;
		digit--;
		stemp /= 10;
	} while (stemp > 0);

	while (digit > 0) {
		map[x + y*32 + (digit-1)] = DIGIT_CHIP + pal * 4096;
		digit--;
	}
	return;
}
void 
PrintDigit(u16 *map, const s32 num, u8 digit, s16 x, s16 y, u8 pal)
{
	s32 stemp = num;
	if (num < 0) {
		stemp *= -1;
		map[x + y*32 - 1] = 13 + pal * 4096;
	}
	if (stemp == 0) {
		map[x + y*32 + (digit-1)] = 0 + DIGIT_CHIP + pal * 4096;
		digit--;
	} else {
		while (stemp > 0) {
			int number = (stemp & 0xf);
			if (number > 9) number += ALPHABET_CHIP - 10;
			else number += DIGIT_CHIP;
			map[x + y*32 + (digit-1)] = number + pal * 4096;
			if (digit == 0) break;
			digit--;
			stemp >>= 4;
		}
	}

	while (digit > 0) {
		map[x + y*32 + (digit-1)] = 0x00 + pal * 4096;
		digit--;
	}
	return;
}
void 
PrintHex(u16 *map, const u32 num, u8 digit, s16 x, s16 y, u8 pal)
{
	u32 stemp = num;
	if (stemp == 0) {
		map[x + y*32 + (digit-1)] = 0 + HEX_CHIP + pal * 4096;
		digit--;
	} else {
		while (stemp > 0) {
			int number = (stemp & 0xf);
			number += HEX_CHIP;
			map[x + y*32 + (digit-1)] = number + pal * 4096;
			if (digit == 0) break;
			digit--;
			stemp >>= 4;
		}
	}

	while (digit > 0) {
		map[x + y*32 + (digit-1)] = HEX_CHIP + pal * 4096;
		digit--;
	}
	return;
}
