//==================================================================
// gba_define.h
//==================================================================
#ifndef GBA_DEFINE_H
#define GBA_DEFINE_H

//--------------------------------------------------------------
// defines
//--------------------------------------------------------------
#define ABS(x)	( ((x) >= 0) ? (x) : (-(x)) )
#define SIGN(x)	( ((x) >= 0) ? (1) : (-(1)) )

#define	BOOL	char
#define	TRUE	1
#define	FALSE	0
#define	NULL	((void*)0)

#define BIT0	(0x0001)
#define BIT1	(0x0002)
#define BIT2	(0x0004)
#define BIT3	(0x0008)
#define BIT4	(0x0010)
#define BIT5	(0x0020)
#define BIT6	(0x0040)
#define BIT7	(0x0080)
#define BIT8	(0x0100)
#define BIT9	(0x0200)
#define BIT10	(0x0400)
#define BIT11	(0x0800)
#define BIT12	(0x1000)
#define BIT13	(0x2000)
#define BIT14	(0x4000)
#define BIT15	(0x8000)
#define BIT16	(0x00010000)
#define BIT17	(0x00020000)
#define BIT18	(0x00040000)
#define BIT19	(0x00080000)
#define BIT20	(0x00100000)
#define BIT21	(0x00200000)
#define BIT22	(0x00400000)
#define BIT23	(0x00800000)
#define BIT24	(0x01000000)
#define BIT25	(0x02000000)
#define BIT26	(0x04000000)
#define BIT27	(0x08000000)
#define BIT28	(0x10000000)
#define BIT29	(0x20000000)
#define BIT30	(0x40000000)
#define BIT31	(0x80000000)

#define FIX_POWER		(0x100)					// Œ菬_ 8rbg
#define INT2F( _x )		( (_x) * FIX_POWER )
#define FIX2I( _x )		( (_x) / FIX_POWER )

//--------------------------------------------------------------
// Function Attributes
//--------------------------------------------------------------
#define IN_IWRAM __attribute__ ((section (".iwram")))
#define IN_EWRAM __attribute__ ((section (".ewram")))
#ifndef DEBUG
#	define CODE_IN_IWRAM __attribute__ ((section (".iwram"), long_call))
#	define CODE_IN_EWRAM __attribute__ ((section (".ewram"), long_call))
#	define CODE_IN_IWRAM_NAKED __attribute__ ((section (".iwram"), long_call, naked))
#	define CODE_IN_EWRAM_NAKED __attribute__ ((section (".ewram"), long_call, naked))
#else 
#	define CODE_IN_IWRAM 
#	define CODE_IN_EWRAM 
#	define CODE_IN_IWRAM_NAKED __attribute__ ((naked))
#	define CODE_IN_EWRAM_NAKED __attribute__ ((naked))
#endif
#define ATTR_PACKED __attribute__ ((packed))
#define ATTR_NAKED __attribute__ ((naked))
#define ATTR_IRQ __attribute__ ((interrupt ("IRQ")))


//--------------------------------------------------------------
// DispControl Flags
//--------------------------------------------------------------
#define DISP_MODE_0             0x0000 // BG Mode 0 
#define DISP_MODE_1             0x0001 // BG Mode 1
#define DISP_MODE_2             0x0002 // BG Mode 2
#define DISP_MODE_3             0x0003 // BG Mode 3
#define DISP_MODE_4             0x0004 // BG Mode 4
#define DISP_MODE_5             0x0005 // BG Mode 5
#define DISP_BMP_FRAME_0        0x0000 // Bitmap Frame Select 0
#define DISP_BMP_FRAME_1        0x0010 // Bitmap Frame Select 1
#define DISP_OAM_HBLANK         0x0020 // Force processing during HBLANK
#define DISP_OBJ_2D_MAP         0x0000 // OBJ Mapping Type 2D
#define DISP_OBJ_1D_MAP         0x0040 // OBJ Mapping Type 1D
#define DISP_FORCE_BLANK        0x0080 // Force the display to go Blank
#define DISP_BG0_ON             0x0100 // BG0 On
#define DISP_BG1_ON             0x0200 // BG1 On
#define DISP_BG2_ON             0x0400 // BG2 On
#define DISP_BG3_ON             0x0800 // BG3 On
#define DISP_OBJ_ON             0x1000 // OBJ ON
#define DISP_WIN0_ON            0x2000 // Window 0 On
#define DISP_WIN1_ON            0x4000 // Window 1 On
#define DISP_OBJWIN_ON          0x8000 // OBJ Window On

#define DISP_BG_CHAR_OFS(x)     (((x) & 0x1f) << 8)
#define DISP_BG_SCREEN_OFS(x)   (((x) & 0x03) << 2)

//--------------------------------------------------------------
// DispStatus Flags
//--------------------------------------------------------------
#define VCOUNT_IF_MASK          0xFF80
#define VCOUNT_IF_SHIFT         0x07
#define STAT_MASK               0x007f

#define STAT_VBLANK             0x0001
#define STAT_HBLANK             0x0002
#define STAT_VCOUNT             0x0004
#define STAT_VBLANK_IF          0x0008
#define STAT_HBLANK_IF          0x0010
#define STAT_VCOUNT_IF          0x0020

//--------------------------------------------------------------
// Window Flags
//--------------------------------------------------------------
#define WIN_SHIFT               0x08

#define WIN_BG0_ON              0x0001 // BG0 ON
#define WIN_BG1_ON              0x0002 // BG1 ON
#define WIN_BG2_ON              0x0004 // BG2 ON
#define WIN_BG3_ON              0x0008 // BG3 ON
#define WIN_OBJ_ON              0x0010 // OBJ ON
#define WIN_BLEND_ON            0x0020 // Blend ON

//--------------------------------------------------------------
// Timer Control Flags
//--------------------------------------------------------------
#define TIMER_PRESCALER_MASK    0x0003
#define TIMER_PRESCALER_SHIFT   16

#define TIMER_PRESCALER_1CLK    0x0000
#define TIMER_PRESCALER_64CLK   0x0001
#define TIMER_PRESCALER_256CLK  0x0002
#define TIMER_PRESCALER_1024CK  0x0003
#define TIMER_FLOWCOUNT         0x0004
#define TIMER_IF_ENABLE         0x0040
#define TIMER_ENABLE            0x0080
#define TIMER_DISABLE           0x0000

//--------------------------------------------------------------
// DMA Control Flags
//--------------------------------------------------------------
#define DMA0COPY(sad,dad,size,mode)	\
	REG_DMA0SAD = (sad);		\
	REG_DMA0DAD = (dad);		\
	REG_DMA0CNT = (size) + (mode);

#define DMA1COPY(sad,dad,size,mode)	\
	REG_DMA1SAD = (sad);		\
	REG_DMA1DAD = (dad);		\
	REG_DMA1CNT = (size) + (mode);

#define DMA2COPY(sad,dad,size,mode)	\
	REG_DMA2SAD = (sad);		\
	REG_DMA2DAD = (dad);		\
	REG_DMA2CNT = (size) + (mode);

#define DMA3COPY(sad,dad,size,mode)	\
	REG_DMA3SAD = (sad);		\
	REG_DMA3DAD = (dad);		\
	REG_DMA3CNT = (size) + (mode);

#define	DMA_DEST_INC            0x00000000 // Destination Adress Increment
#define	DMA_DEST_DEC            0x00200000 // Destination Adress Decriment
#define	DMA_DEST_FIX            0x00400000 // Destination Adress Fixed
#define	DMA_DEST_RELOAD         0x00600000 // Destination Adress Increment & Reload
#define	DMA_SRC_INC             0x00000000 // Source Increment
#define	DMA_SRC_DEC             0x00800000 // Source Decriment
#define	DMA_SRC_FIX             0x01000000 // Source Fixed
#define	DMA_REPEAT              0x02000000 // DMA Enable when transfer finished (DMA_V/HBlank bit required)
#define	DMA_16BIT               0x00000000 // 16Bit Bus
#define	DMA_32BIT               0x04000000 // 32Bit Bus
#define	DMA_IMMEDIATE           0x00000000 // Start Immediately
#define	DMA_CARTREQ             0x08000000 // When Cart Request
#define	DMA_VBLANK              0x10000000 // When VBlank
#define	DMA_HBLANK              0x20000000 // When HBlank
#define DMA_LINE                0x30000000 // When Line displayed (BG-Bitmap)
#define DMA_FIFO                0x30000000 // When FIFO Requested
#define	DMA_IF_ENABLE           0x40000000 // Interrupt Enable
#define	DMA_ENABLE              0x80000000 // DMA Enable

//--------------------------------------------------------------
// Game Pak Wait State Control Flags
//--------------------------------------------------------------
#define WS_SRAM_4WAIT           0x0000
#define WS_SRAM_3WAIT           0x0001
#define WS_SRAM_2WAIT           0x0002
#define WS_SRAM_8WAIT           0x0003
#define WS_ROM0_1ST_4WAIT       0x0000
#define WS_ROM0_1ST_3WAIT       0x0004
#define WS_ROM0_1ST_2WAIT       0x0008
#define WS_ROM0_1ST_8WAIT       0x000C
#define WS_ROM0_2ND_2WAIT       0x0000
#define WS_ROM0_2ND_1WAIT       0x0010
#define WS_ROM1_1ST_4WAIT       0x0000
#define WS_ROM1_1ST_3WAIT       0x0020
#define WS_ROM1_1ST_2WAIT       0x0040
#define WS_ROM1_1ST_8WAIT       0x0060
#define WS_ROM1_2ND_4WAIT       0x0000
#define WS_ROM1_2ND_1WAIT       0x0080
#define WS_ROM2_1ST_4WAIT       0x0000
#define WS_ROM2_1ST_3WAIT       0x0100
#define WS_ROM2_1ST_2WAIT       0x0200
#define WS_ROM2_1ST_8WAIT       0x0300
#define WS_ROM2_2ND_8WAIT       0x0000
#define WS_ROM2_2ND_1WAIT       0x0400
#define WS_PREFETCH_ENABLE      0x4000
#define WS_MODE_AGB             0x0000
#define WS_MODE_CGB             0x0000

#define WS_FASTEST              (WS_SRAM_2WAIT \
                                | WS_ROM0_1ST_2WAIT | WS_ROM0_2ND_1WAIT \
                                | WS_ROM1_1ST_2WAIT | WS_ROM1_2ND_1WAIT \
                                | WS_ROM2_1ST_2WAIT | WS_ROM2_2ND_1WAIT \
                                | WS_MODE_AGB | WS_PREFETCH_ENABLE)
#define WS_FASTER               (WS_SRAM_3WAIT \
                                | WS_ROM0_1ST_3WAIT | WS_ROM0_2ND_1WAIT \
                                | WS_ROM1_1ST_3WAIT | WS_ROM1_2ND_1WAIT \
                                | WS_ROM2_1ST_3WAIT | WS_ROM2_2ND_1WAIT \
                                | WS_MODE_AGB | WS_PREFETCH_ENABLE)
#define WS_NORMAL               (WS_SRAM_8WAIT \
                                | WS_ROM0_1ST_3WAIT | WS_ROM0_2ND_1WAIT \
                                | WS_ROM1_1ST_3WAIT | WS_ROM1_2ND_1WAIT \
                                | WS_ROM2_1ST_3WAIT | WS_ROM2_2ND_1WAIT \
                                | WS_MODE_AGB | WS_PREFETCH_ENABLE)

//--------------------------------------------------------------
// Interrupt Control Flags
//--------------------------------------------------------------
#define INTR_VBLANK             0x0001
#define INTR_HBLANK             0x0002
#define INTR_VCOUNT             0x0004
#define INTR_TIMER0             0x0008
#define INTR_TIMER1             0x0010
#define INTR_TIMER2             0x0020
#define INTR_TIMER3             0x0040
#define INTR_SIO                0x0080
#define INTR_DMA0               0x0100
#define INTR_DMA1               0x0200
#define INTR_DMA2               0x0400
#define INTR_DMA3               0x0800
#define INTR_KEY                0x1000
#define INTR_GBA_CART           0x2000

#define IME_START               0x0001
#define IME_STOP                0x0000

//--------------------------------------------------------------
// Blend Control Flags
//--------------------------------------------------------------
#define BLEND_1ST_MASK          0x003f
#define BLEND_2ND_MASK          0x3f00
#define BLEND_MODE_MASK         0x00c0

#define BLEND_1ST_SHIFT         0
#define BLEND_2ND_SHIFT         8
#define BLEND_MODE_SHIFT        6

#define BLEND_BG0               0x0001 // BG0 Pixel
#define BLEND_BG1               0x0002 // BG1 Pixel
#define BLEND_BG2               0x0004 // BG2 Pixel
#define BLEND_BG3               0x0008 // BG3 Pixel
#define BLEND_OBJ               0x0010 // OBJ Pixel
#define BLEND_BD                0x0020 // Backdrop Pixel
#define BLEND_ALL               0x003f // All Pixels
#define BLEND_NONE              0x0000 // 
#define BLEND_ALPHA             0x0040 // 1st + 2nd
#define BLEND_BRIGHT            0x0080 // 1st becomes brighter
#define BLEND_DARK              0x00C0 // 1st becomes darker

#define BLEND_1ST_BG0           0x0001 // 1st BG0 Pixel
#define BLEND_1ST_BG1           0x0002 // 1st BG1 Pixel
#define BLEND_1ST_BG2           0x0004 // 1st BG2 Pixel
#define BLEND_1ST_BG3           0x0008 // 1st BG3 Pixel
#define BLEND_1ST_OBJ           0x0010 // 1st OBJ Pixel
#define BLEND_1ST_BD            0x0020 // 1st Backdrop Pixel
#define BLEND_1ST_ALL           0x003f // 1st All Pixels
#define BLEND_2ND_BG0           0x0100 // 2nd BG0 Pixel
#define BLEND_2ND_BG1           0x0200 // 2nd BG1 Pixel
#define BLEND_2ND_BG2           0x0400 // 2nd BG2 Pixel
#define BLEND_2ND_BG3           0x0800 // 2nd BG3 Pixel
#define BLEND_2ND_OBJ           0x1000 // 2nd OBJ Pixel
#define BLEND_2ND_BD            0x2000 // 2nd Backdrop Pixel
#define BLEND_2ND_ALL           0x3f00 // 2nd All Pixels

#define COLEV_EVA_MASK          0x001f
#define COLEV_EVB_MASK          0x1f00
#define COLY_MASK               0x001f

#ifndef BLEND32
#	define COLEV_EVA_SHIFT         0
#	define COLEV_EVB_SHIFT         8
#	define COLY_SHIFT              0
#else
#	define COLEV_EVA_SHIFT         16
#	define COLEV_EVB_SHIFT         24
#	define COLY_SHIFT              0
#endif

//--------------------------------------------------------------
// Sound Control Flags
//--------------------------------------------------------------

// REG_SGCNT0_L
#define SG0_VOLUME_R_SHIFT      0
#define SG0_VOLUME_L_SHIFT      4
#define SG0_VOLUME_R_MASK       0x0003
#define SG0_VOLUME_L_MASK       0x0030

#define SG0_SOUND_1_R_ON         0x0100
#define SG0_SOUND_2_R_ON         0x0200
#define SG0_SOUND_3_R_ON         0x0400
#define SG0_SOUND_4_R_ON         0x0800
#define SG0_SOUND_1_L_ON         0x1000
#define SG0_SOUND_2_L_ON         0x2000
#define SG0_SOUND_3_L_ON         0x4000
#define SG0_SOUND_4_L_ON         0x8000

// REG_SGCNT0_H
#define SG0_SOUND_OUT_MASK      0x0003
#define SG0_DSOUND_A_OUT_MASK   0x0004
#define SG0_DSOUND_B_OUT_MASK   0x0008

#define SG0_SOUND_OUT_QUATER    0x0000
#define SG0_SOUND_OUT_HALF      0x0001
#define SG0_SOUND_OUT_FULL      0x0002
#define SG0_DSOUND_A_OUT_HALF   0x0000
#define SG0_DSOUND_A_OUT_FULL   0x0004
#define SG0_DSOUND_B_OUT_HALF   0x0000
#define SG0_DSOUND_B_OUT_FULL   0x0008

#define SG0_DSOUND_A_R_ON       0x0100
#define SG0_DSOUND_A_L_ON       0x0200
#define SG0_DSOUND_A_TIMER0     0x0000
#define SG0_DSOUND_A_TIMER1     0x0400
#define SG0_DSOUND_A_RESET      0x0800
#define SG0_DSOUND_B_R_ON       0x1000
#define SG0_DSOUND_B_L_ON       0x2000
#define SG0_DSOUND_B_TIMER0     0x0000
#define SG0_DSOUND_B_TIMER1     0x4000
#define SG0_DSOUND_B_RESET      0x8000

#define SG1_SOUND_1_ON	        0x0001
#define SG1_SOUND_2_ON	        0x0002
#define SG1_SOUND_3_ON	        0x0004
#define SG1_SOUND_4_ON	        0x0008
#define SG1_SOUND_ALL_ON	    0x0080

#endif // GBA_DEFINE_H
