//==============================================================================
// sprite.h
//==============================================================================
#ifndef _SPRITE_H
#define _SPRITE_H
#include <gba.h>
#ifdef __cplusplus
extern "C" {
#endif

// #define USE_SPR_MACRO
// #define USE_DOUBLER

//--------------------------------------------------------------------
// őgpXvCgύXꍇ MAX_WORKSPR Ă
//--------------------------------------------------------------------
#define MAX_WORKSPR			(128)

//----------------------------------
#define MAX_TILES			(1024)
#define MAX_SPRITES			(128)
#define INVALID_SPR			(255)


// Atribute0 -----------------------
#define SP_ROTATION_FLAG 	0x0100
#define SP_SIZE_DOUBLE		0x0200
#define SP_MODE_NORMAL     	0x0000
#define SP_MODE_TRANSPERANT	0x0400
#define SP_MODE_WINDOWED	0x0800
#define SP_MOSAIC			0x1000
#define SP_COLOR_16			0x0000
#define SP_COLOR_256		0x2000
#define SP_SQUARE			0x0000
#define SP_WIDE				0x4000
#define SP_TALL				0x8000
// Atribute1 -----------------------
#define SP_ROTDATA(n)		((n) << 9)
#define SP_HORIZONTAL_FLIP	0x1000
#define SP_VERTICAL_FLIP	0x2000
#define SP_SIZE_8			0x0000
#define SP_SIZE_16			0x4000
#define SP_SIZE_32			0x8000
#define SP_SIZE_64			0xC000
// Atribute2 -----------------------
#define SP_PRIORITY(n)		((n) << 10)
#define SP_PALETTE(n)		((n) << 12)

#define GET_OBJ_SHAPE(x)	( ((((x)->attr0) & 0xC000) >> 12) + ((((x)->attr1) & 0xC000) >> 14) )
#define OBJ_SIZE(x)		SprCopySize[( x )]
#define OBJ_BLOCK(x)	SprBlockSize[( x )]
#define OBJ_V_LENGTH(x)	SprVerticalLength[( x )]
#define OBJ_H_LENGTH(x)	SprHorizontalLength[( x )]

// 
#define OBJ_SIZE_8x8	0
#define OBJ_SIZE_16x16	1
#define OBJ_SIZE_32x32	2
#define OBJ_SIZE_64x64	3
#define OBJ_SIZE_16x8	4
#define OBJ_SIZE_32x8	5
#define OBJ_SIZE_32x16	6
#define OBJ_SIZE_64x32	7
#define OBJ_SIZE_8x16	8
#define OBJ_SIZE_8x32	9
#define OBJ_SIZE_16x32	10
#define OBJ_SIZE_32x64	11
// u16P
#define OBJ_SHORT_SIZE_8x8		16
#define OBJ_SHORT_SIZE_16x16	64
#define OBJ_SHORT_SIZE_32x32	256
#define OBJ_SHORT_SIZE_64x64	1024
#define OBJ_SHORT_SIZE_16x8		32
#define OBJ_SHORT_SIZE_32x8		64
#define OBJ_SHORT_SIZE_32x16	128
#define OBJ_SHORT_SIZE_64x32	512
#define OBJ_SHORT_SIZE_8x16		32
#define OBJ_SHORT_SIZE_8x32		64
#define OBJ_SHORT_SIZE_16x32	128
#define OBJ_SHORT_SIZE_32x64	512
// u16 * 16 P
#define OBJ_BLOCK_SIZE_8x8		1
#define OBJ_BLOCK_SIZE_16x16	4
#define OBJ_BLOCK_SIZE_32x32	16
#define OBJ_BLOCK_SIZE_64x64	64
#define OBJ_BLOCK_SIZE_16x8		2
#define OBJ_BLOCK_SIZE_32x8		4
#define OBJ_BLOCK_SIZE_32x16	8
#define OBJ_BLOCK_SIZE_64x32	32
#define OBJ_BLOCK_SIZE_8x16		2
#define OBJ_BLOCK_SIZE_8x32		4
#define OBJ_BLOCK_SIZE_16x32	8
#define OBJ_BLOCK_SIZE_32x64	32

//----------------------------------
typedef struct {
	u16 a0;
	u16 a1;
} ATTR_PACKED SPR_SHAPE;

//----------------------------------
typedef struct {
	u16 *addr;
	u8 shape;
	u8 pal; // not used
} ATTR_PACKED SPR_PAT;

//----------------------------------
typedef struct {
	u16 tile[16];
} ATTR_PACKED TILE_BLOCK;

//----------------------------------
typedef struct RotData
{
	u16 filler1[3];
	u16 pa;
	u16 filler2[3];
	u16 pb;	
	u16 filler3[3];
	u16 pc;	
	u16 filler4[3];
	u16 pd;
} RotData, *pRotData;

typedef struct OAMEntry
{
	u16 attribute0;
	u16 attribute1;
	u16 attribute2;
	u16 attribute3;
} OAMEntry, *pOAMEntry;

typedef struct OAMEntry32
{
	u32 attribute01;
	u16 attribute2;
	u16 attribute3;
} OAMEntry32, *pOAMEntry32;

typedef struct SPRWORK {
	struct SPRWORK *prev;		// ỎzXvCg
	struct SPRWORK *next;		// ̉zXvCg
	u8	disp;		// \tO
	u8	priority;	// D揇
	u16	attr0;
	u16	attr1;
	u16	attr2;
} __attribute__ ((packed)) SPRWORK;

typedef struct SPRWORK32 {
	struct SPRWORK32 *prev;		// ỎzXvCg
	struct SPRWORK32 *next;		// ̉zXvCg
	u8	disp;		// \tO
	u8	priority;	// D揇
	u32	attr01;
	u16	attr2;
} __attribute__ ((packed)) SPRWORK32;

//----------------------------------

extern const u16 SprBlockSize[12];
extern const u16 SprCopySize[12];
extern const u16 SprHorizontalLength[12];
extern const u16 SprVerticalLength[12];
extern const SPR_SHAPE SprShapeAttribute[12];
extern SPRWORK tempSprWork;

//----------------------------------

CODE_IN_IWRAM void SpriteInit( SPRWORK* , SPR_PAT* );
CODE_IN_IWRAM SPRWORK *SpriteSet( u8 );
CODE_IN_IWRAM void SpriteKill( SPRWORK* );
CODE_IN_IWRAM void SpriteUpdate( void );
CODE_IN_IWRAM void SpritePriority( SPRWORK* , u8 );

#ifndef USE_SPR_MACRO

CODE_IN_IWRAM void SpriteSetWorld(s32 x, s32 y);
CODE_IN_IWRAM void SpriteSetView(s32 x, s32 y);

CODE_IN_IWRAM void SpritePattern( SPRWORK* work, u16 pattern );
CODE_IN_IWRAM void SpriteMove( SPRWORK*, s16, s16 );
CODE_IN_IWRAM void SpriteMoveWorld( SPRWORK*, s16, s16 );
CODE_IN_IWRAM void SpriteSize( SPRWORK*, u16 );
CODE_IN_IWRAM void SpriteChangeNumber( SPRWORK*, u16 );
CODE_IN_IWRAM void SpriteChangePallete( SPRWORK*, u16 );
CODE_IN_IWRAM void SpriteChangePriority( SPRWORK*, u16 num );
CODE_IN_IWRAM void SpriteChangeFlip( SPRWORK*, u16 num );
CODE_IN_IWRAM void SpriteMode(SPRWORK* work, u16 num);

#else // USE_SPR_MACRO

#include <pcgcache.h>
extern SPR_PAT *SprPattern;

//---------------------------------------------------------
// XvCgLN^ꊇύX
//---------------------------------------------------------
#define SpritePattern(work, num) {\
{\
	SPR_PAT *_pat = &SprPattern[(num)];\
	u16 a0, a1, tile, old;\
\
	if (work == &tempSprWork) return;\
\
	/* `𓾂 */\
	a0 = SprShapeAttribute[_pat->shape].a0;\
	a1 = SprShapeAttribute[_pat->shape].a1;\
\
	/* p^[ */\
	old = work->attr2 & 0x03ff;\
	PCG_ReleaseTile(old);\
\
	/* ^Cԍ𓾂 */\
	tile = PCG_Set(num);\
\
	work->attr0 &= ~(SP_WIDE | SP_TALL);\
	work->attr0 |= a0;\
	work->attr1 &= ~(SP_SIZE_8 | SP_SIZE_16 | SP_SIZE_32 | SP_SIZE_64);\
	work->attr1 |= a1;\
	work->attr2 &= ~(MAX_TILES - 1);\
	work->attr2 |= tile;\
}\
}

//---------------------------------------------------------
// XvCgړ
//---------------------------------------------------------
#define SpriteMove(work, x, y) {\
{\
	s16 xx = (x);\
	s16 yy = (y);\
	u16 shape = GET_OBJ_SHAPE(work);\
	if (xx > 240		/* X\Oꂽ */\
	    || yy > 160		/* Y\Oꂽ */\
	    || (xx < - SprHorizontalLength[shape] )	/* X\Oꂽ */\
	    || (yy < - SprVerticalLength[shape] )	/* Y\Oꂽ */\
	   )\
	{\
		xx = 240;\
		yy = 160;\
	} else { \
		if (xx < 0) xx = 512 + xx;\
		if (yy < 0) yy = 256 + yy;\
	}\
\
	work->attr1 = work->attr1 & 0xFC00;\
	work->attr1 = work->attr1 | xx;\
	work->attr0 = work->attr0 & 0xFF00;\
	work->attr0 = work->attr0 | yy;\
}\
}

//---------------------------------------------------------
//	XvCg̃TCYݒ
//---------------------------------------------------------
#define SpriteSize(work, size) {\
{\
	work->attr1 &= 0x3FFF;\
	work->attr1 |= ((size & 3) << 14);\
}\
}

//---------------------------------------------------------
// XvCgLN^ύX
//---------------------------------------------------------
#define SpriteChangeNumber(work, num) {\
{\
	u16 p = num;\
	p &= 0x03ff;\
	work->attr2 &= 0xFC00;\
	work->attr2 |= p;\
}\
}

//---------------------------------------------------------
// XvCgpbgύX
//---------------------------------------------------------
#define SpriteChangePallete(work, num) {\
{\
	u16 p = num;\
	p &= 0x0f;\
	p <<= 12;\
	work->attr2 &= 0x0fff;\
	work->attr2 |= p;\
}\
}

//---------------------------------------------------------
// XvCgvCIeBύX
//---------------------------------------------------------
#define SpriteChangePriority(work, num) {\
{\
	u16 p = num;\
	p &= 0x03;\
	p <<= 10;\
	work->attr2 &= 0xf3ff;\
	work->attr2 |= p;\
}\
}

//---------------------------------------------------------
// XvCgtbvύX
//---------------------------------------------------------
#define SpriteChangeFlip(work, num) {\
{\
	u16 p = num;\
	p &= (SP_HORIZONTAL_FLIP | SP_VERTICAL_FLIP);\
	work->attr1 &= ~(SP_HORIZONTAL_FLIP | SP_VERTICAL_FLIP);\
	work->attr1 |= p;\
}\
}

//---------------------------------------------------------
// XvCg[hύX
//---------------------------------------------------------
#define SpriteMode(work, num) {\
{\
	u16 p = num;\
	p &= (SP_MODE_TRANSPERANT | SP_MODE_WINDOWED);\
	work->attr0 &= ~(SP_MODE_TRANSPERANT | SP_MODE_WINDOWED);\
	work->attr0 |= p;\
}\
}

#endif // USE_SPR_MACRO


#ifdef __cplusplus
}
#endif
#endif // _SPRITE_H
