#include <isr.h>
#include <key.h>
#include <dmachain.h>
#include <pcgcache.h>
#include <sprite.h>
#include <gamemath.h>
//#include "dft.h"
#include "disp.h"
#include "pdrv.h"
#include "nmid.h"
#include "res/_ChrTable.h"

extern u8 akuden;
extern u8 U4_Wanderer;
extern u8 fantasie0;
extern u8 etude10_12;
extern u8 garegga1st_agb;
extern u8 fantasie_3tone;

#define REPEAT_TIME		15
#define REPEAT_SPEED	4
#define REPEAT_DELAY	(REPEAT_TIME - REPEAT_SPEED)

//--------------------------------------------------------------
// Globals
SPRWORK sprwork[MAX_WORKSPR] IN_EWRAM;
s8 pcm_buffer[2][712] IN_EWRAM;
//s32 dftbuf[2][712];

extern PDRVWORK pdrvw;
extern INSTBANK inst[];
static u16 rateNum = 3;
// static volatile int dft_enable = 1;

static const struct {
	u8* song;
	s8* title;
} SongConstants[] = {
	{ &garegga1st_agb, "BATTLE GAREGGA - Stage 1", },
	{ &akuden, "CastlevaniaIII - Prelude", },
	{ &fantasie_3tone, "Fantasie 3tones", },
	{ &fantasie0, "Chopin - Fantasie", },
	{ &etude10_12, "Chopin - Revolution", },
	{ &U4_Wanderer, "UltimaIV - Wanderer", },
};

struct {
	u8 max;
	u8 num;
} song;

struct {
	u8 cur;
	u8 old;
} cursor;

u32		bpm = 0x100;
u32		time = 0;
u8		note = 0;
s16		pitch = 0;
u8		volume = 128;
u16		freq;
u32		cpu = 0;
vu32	swt = 1;
u16		buflen;
u16 mute[16] = {
	0,0,0,0,0,0,0,0,
	0,0,0,0,0,0,0,0,
};

//--------------------------------------------------------------
//
void CursorAction(u8 num, u16 factor)
{
	int i;
	switch (num) {
	case CUR_TRK_0:
	case CUR_TRK_1:
	case CUR_TRK_2:
	case CUR_TRK_3:
	case CUR_TRK_4:
	case CUR_TRK_5:
	case CUR_TRK_6:
	case CUR_TRK_7:
	case CUR_TRK_8:
	case CUR_TRK_9:
	case CUR_TRK_A:
	case CUR_TRK_B:
	case CUR_TRK_C:
	case CUR_TRK_D:
	case CUR_TRK_E:
	case CUR_TRK_F:
		if (factor == KEY_ABUTTON) {
			num = num - CUR_TRK_0;
			mute[num] = 1 - mute[num];
			NMID_Mute(num, mute[num]);
		} else {
			for (i = 0; i < NMID_MAX_TRACKS; i++) {
				mute[i] = 1 - mute[i];
				NMID_Mute(i, mute[i]);
			}
		}
		break;
	case CUR_VOL:
		if (factor == KEY_BBUTTON) {
			if (pdrvw.master_vol > 0) pdrvw.master_vol-=2;
	    } else {
			if (pdrvw.master_vol < 0x80) pdrvw.master_vol+=2;
		}
		PrintHex(map0, pdrvw.master_vol, 2, 2, 18, 1);
		break;
	case CUR_KEY:
		if (factor == KEY_BBUTTON) {
			if (pdrvw.master_keyshift > -127) pdrvw.master_keyshift--;
		} else {
			if (pdrvw.master_keyshift < 127) pdrvw.master_keyshift++;
		}
		PrintHex(map0, pdrvw.master_keyshift & 0xff, 2, 6, 18, 1);
		break;
	case CUR_BACK:
		if (factor == KEY_ABUTTON) {
			NMID_SongSet(SongConstants[song.num].song);
			time = 0;
		}
		break;
	default:
		break;
	}
}

//--------------------------------------------------------------
void VsyncHandler(void)
{
	static u32 keyrepeat;
	u32 cpu_count;

	PDRV_Switch(&pcm_buffer[swt][0]);
	swt = 1 -swt;

	DMAChain_Transfer();
	SpriteUpdate();
	KeyCheck();

	if (key.cnt != 0 && key.cnt == key.old) {
		keyrepeat++;
	} else {
		keyrepeat = 0;
	}

	if ( key.trg & KEY_UP || (key.cnt & KEY_UP && keyrepeat > REPEAT_TIME) ) {
		if (cursor.cur > 0) {
			cursor.cur--;
			PrintCursor(cursor.old, 0);
			PrintCursor(cursor.cur, 1);
			cursor.old = cursor.cur;
		}
		if (keyrepeat > REPEAT_TIME)
			keyrepeat = REPEAT_DELAY;
	} else 
	if ( key.trg & KEY_DOWN || (key.cnt & KEY_DOWN && keyrepeat > REPEAT_TIME) ) {
		if (cursor.cur < CUR_NONE - 1) {
			cursor.cur++;
			PrintCursor(cursor.old, 0);
			PrintCursor(cursor.cur, 1);
			cursor.old = cursor.cur;
		}
		if (keyrepeat > REPEAT_TIME)
			keyrepeat = REPEAT_DELAY;
	} else {
		if (cursor.cur == CUR_BPM)
			PrintDigit0I(map0, nmidi.bpm, 3, 10, 18, 1);
		else
			PrintDigit0I(map0, nmidi.bpm, 3, 10, 18, 0);
	}


	if (key.trg & KEY_RIGHT) {
		song.num++;
		if (song.num >= song.max)
			song.num = 0;
		NMID_SongSet(SongConstants[song.num].song);
		PrintStr(map0, "                             ", 0, 19, 1);
		PrintStr(map0, SongConstants[song.num].title, 0, 19, 1);
		time = 0;
	} else 
	if (key.trg & KEY_LEFT) {
		if (song.num == 0)
			song.num = song.max;
		song.num--;
		NMID_SongSet(SongConstants[song.num].song);
		PrintStr(map0, "                             ", 0, 19, 1);
		PrintStr(map0, SongConstants[song.num].title, 0, 19, 1);
		time = 0;
	}

	if (key.trg & KEY_RBUTTON) {
		if (cursor.cur > CUR_BPM) {
			cursor.cur = CUR_TRK_0;
		} else
		if (cursor.cur > CUR_TRK_F) {
			cursor.cur = CUR_BPM + 1;
		} else {
			cursor.cur = CUR_TRK_F + 1;
		}
		PrintCursor(cursor.old, 0);
		PrintCursor(cursor.cur, 1);
		cursor.old = cursor.cur;
	}

	if (key.trg & KEY_ABUTTON
		|| (key.cnt & KEY_ABUTTON 
			&& keyrepeat > REPEAT_TIME 
			&& cursor.cur > CUR_TRK_F) ) {
		CursorAction(cursor.cur, KEY_ABUTTON);
		if (keyrepeat > REPEAT_TIME)
			keyrepeat = REPEAT_DELAY;
	}

	if (key.trg & KEY_BBUTTON
		|| (key.cnt & KEY_BBUTTON 
			&& keyrepeat > REPEAT_TIME 
			&& cursor.cur > CUR_TRK_F) ) {
		CursorAction(cursor.cur, KEY_BBUTTON);
		if (keyrepeat > REPEAT_TIME)
			keyrepeat = REPEAT_DELAY;
	}

	if (key.trg & KEY_START) {
		if (rateNum < 9) {
			rateNum++;
			buflen = PDRV_SetRate(rateNum);
		}
		PrintRate(rateNum);
	} else
	if (key.trg & KEY_SELECT) {
		if (rateNum > 0) {
			rateNum--;
			buflen = PDRV_SetRate(rateNum);
		}
		PrintRate(rateNum);
	}

	PDRV_Render(buflen, &pcm_buffer[swt][0]);
	NMID_Play();

	cpu_count = REG_VCOUNT;
	TrackDispUpdate();
	if (cpu_count >= 160) {
		cpu_count -= 160;
	} else {
		cpu_count += 68;
	}
	cpu_count = FIXDIV(cpu_count * 100, 228);
	cpu_count = (cpu_count + cpu) / 2;
	cpu = cpu_count;
	PrintCpu(cpu_count);
	PrintTime(time);
	time++;
#if 0
	if (dft_enable == 1) {
		dft_enable = 0;
		DFT(&dftbuf[0][0], &dftbuf[1][0], &pcm_buffer[swt][0], rateNum);
	} else {
		dft_enable = 1;
	}
#endif

	return;
}

//--------------------------------------------------------------
// Main
//--------------------------------------------------------------
int main(void)
{
	BOOL done = FALSE;

	REG_WSCNT = WS_FASTER;
	REG_IRQ_VECTOR = (u32)&IRQ_Handler;

	song.max = sizeof(SongConstants) / sizeof(SongConstants[0]);
	song.num = 0;

	cursor.cur = 0;
	cursor.old = 0;
	PrintCursor(cursor.cur, 1);
	PrintRate(rateNum);

	KeyArray_Init();

	PDRV_Init(rateNum, inst);
	PDRV_MasterVolume(0x40);
	NMID_Init();
	NMID_SongSet(SongConstants[song.num].song);

	PCG_Init();
	SpriteInit(sprwork, (SPR_PAT *)&ChrTable[0]);
	DispInit();
	PrintStr(map0, SongConstants[song.num].title, 0, 19, 1);
	PrintRate(rateNum);
	PrintCursor(cursor.cur, 1);

	IRQ_RegistFunc(Int_VBLANK, (u32*)&VsyncHandler);
	REG_DISPSTAT = STAT_VBLANK_IF;

	buflen = PDRV_GetBuf();
	freq = PDRV_GetRate();

	REG_SGCNT0_H = 
		SG0_SOUND_OUT_FULL
		| SG0_DSOUND_A_OUT_FULL
		| SG0_DSOUND_A_R_ON
		| SG0_DSOUND_A_L_ON
		| SG0_DSOUND_A_TIMER0
		| SG0_DSOUND_A_RESET;
	REG_IME = 1;

	while(!done)
	{
//		asm volatile("swi 0x0002");
	}
	return 0;
}

