#include <nmid.h>
#include <nmid_defs.h>
#include <pdrv.h>
#include <gamemath.h>

NMID nmidi IN_EWRAM;
extern PDRVWORK pdrvw;


void NMID_Init(void)
{
	int i, j;

	nmidi.pos = NULL;
	nmidi.pos_start = NULL;
	nmidi.pos_loop = NULL;
	nmidi.count = 0;
	nmidi.channel_mask = 0;
	nmidi.bpm = 120;
#ifdef AGC_ENABLE
	nmidi.agc = 0;
	nmidi.agc_count = 0;
	nmidi.agc_release = 0x10;
#endif
	for (i = 0; i < NMID_MAX_TRACKS; i++) {
		nmidi.status[i].prog = 0;
		nmidi.status[i].vol = 128;
		nmidi.status[i].pan = 64;
		nmidi.status[i].pitch = 0;
		nmidi.status[i].voicemax = NMID_MAX_CHANNELS;
		nmidi.status[i].voicenum = 0;
		nmidi.status[i].porta_speed = 0;
		nmidi.status[i].porta_switch = 0;
		nmidi.status[i].lfo_power = 32;
		nmidi.status[i].lfo_depth = 0;
		nmidi.status[i].lfo_count = 0;
		nmidi.status[i].lfo_speed = 24;
		for (j = 0; j < NMID_MAX_CHANNELS; j++) {
			nmidi.status[i].voice[j].note = 0;
			nmidi.status[i].voice[j].ch = 0;
			nmidi.status[i].voice[j].vel = 128;
			nmidi.status[i].bit.mute = 0;
			nmidi.status[i].bit.monopoly = NMID_TRK_POLY;
		}
	}
	return;
}

void NMID_SongSet(void *pos)
{
	u32 i, j;
	nmidi.pos = pos;
	nmidi.pos_start = pos;
	nmidi.pos_loop = NULL;
	nmidi.count = 0;
	nmidi.bpm = 120;
	for (i = 0; i < NMID_MAX_TRACKS; i++) {
		nmidi.status[i].prog = 0;
		nmidi.status[i].vol = 128;
		nmidi.status[i].pan = 64;
		nmidi.status[i].pitch = 0;
		nmidi.status[i].voicemax = NMID_MAX_CHANNELS;
		nmidi.status[i].voicenum = 0;
		nmidi.status[i].porta_speed = 0;
		nmidi.status[i].porta_switch = 0;
		nmidi.status[i].lfo_power = 32;
		nmidi.status[i].lfo_depth = 0;
		nmidi.status[i].lfo_count = 0;
		nmidi.status[i].lfo_speed = 24;

		if (nmidi.status[i].bit.monopoly != NMID_TRK_POLY) {
			j = nmidi.channel_mask & ~(1 << i);
			nmidi.channel_mask = j;
			nmidi.status[i].bit.monopoly = NMID_TRK_POLY;
		}

		for (j = 0; j < NMID_MAX_CHANNELS; j++) {
			nmidi.status[i].voice[j].note = 0;
			nmidi.status[i].voice[j].ch = 0;
			nmidi.status[i].voice[j].vel = 128;
			nmidi.status[i].voice[j].porta_step = 0;
			nmidi.status[i].voice[j].porta_detune = 0;
		}
	}
	for (i = 0; i < NMID_MAX_CHANNELS; i++) {
		u32 mask = ((nmidi.channel_mask >> i) & 1);
		if (mask)
			continue;
		PDRV_KeyOff(i);
		// force envelope off
//		pdrvw.stat[i].env_phase = ENV_MOD_OFF;
	}
	return;
}

static inline void ControlChange(NMID_DAT *pos)
{
	u8 ch = pos->ch;
	u8 lsb = pos->lsb;
	u8 msb = pos->msb;
	NmidTrackStatus *trk = &nmidi.status[ch];
	u32 i;
	u32 level;

	switch (lsb) {
	// (0x01: Modulation)
	case CC_MODULATION:
		trk->lfo_depth = msb;
		if (msb == 0) {
			trk->lfo_count = 0;
		}
		break;

	// (0x05: Portament Time)
	case CC_PORTAMENT_TIME:
		trk->porta_speed = msb;
		break;

	// (0x07: volume)
	case CC_VOLUME:
		trk->vol = msb;
		for (i = 0; i < trk->voicenum; i++) {
			if (trk->voice[i].note != 0) {
				if (trk->bit.mute) 
					level = 0;
				else
					level = msb * trk->voice[i].vel / 128;
				PDRV_Volume(trk->voice[i].ch, level);
			}
		}
		break;

	// (0x0a: panpot)
	case CC_PANPOT:
		trk->pan = msb - 64;
		break;


	// (0x36: Voice Mode)
	// 0-pcm 1-gb1 2-gb2 3-gb3 4-noise
	case CC_VOICEMODE:
		msb &= 7;
		if (msb > 4) msb = 0;
		trk->bit.mode = msb;
		break;

	// (0x37: Max Voice Nums)
	case CC_VOICENUM:
		trk->voicemax = msb;
		break;

	// (0x38: Loop Point)
	case CC_LOOPPOINT:
		nmidi.pos_loop = pos;
		break;

	// (0x39: Modulation Speed)
	case CC_MODSPEED:
		trk->lfo_speed = msb;
		break;

	// (0x3a: Modulation Power)
	case CC_MODPOWER:
		trk->lfo_power = msb;
		break;

	// (0x41: Portament)
	case CC_PORTAMENT:
		trk->porta_switch = msb;
		break;

	// (0x7c: Omni Off)
	case CC_OMNI_OFF:
		break;
	// (0x7d: Omni On)
	case CC_OMNI_ON:
		break;

	// (0x7e: Mono On)
	case CC_MONO_ON:
		trk->bit.monopoly = NMID_TRK_MONO;
		i = nmidi.channel_mask & ~(1 << ch);
		nmidi.channel_mask = i + (1 << ch);
		break;

	// (0x7f: Poly On)
	case CC_POLY_ON:
		trk->bit.monopoly = NMID_TRK_POLY;
		i = nmidi.channel_mask & ~(1 << ch);
		nmidi.channel_mask = i;
		break;

	default:
		break;
	}
	return;
}

void NMID_Mute(u8 ch, u8 flag)
{
	NmidTrackStatus *trk = &nmidi.status[ch];
	int i;
	u32 level;

	flag &= 1;
	trk->bit.mute = flag;

	for (i = 0; i < trk->voicenum; i++) {
		if (trk->voice[i].note != 0) {
			if (trk->bit.mute)
				level = 0;
			else
				level = trk->vol * trk->voice[i].vel / 128;
			PDRV_Volume(trk->voice[i].ch, level);
		}
	}
}


void NMID_Play(void)
{
	NMID_DAT *pos = nmidi.pos;
	int i, j;

	if (pos == NULL)
		return;

	for (i = 0; i < NMID_MAX_CHANNELS; i++) {
		s16 detune = 0;
		s16 detune_switch = 0;
		NmidTrackStatus *trk = &nmidi.status[i];

		if (trk->lfo_depth != 0) {
			trk->lfo_count = (trk->lfo_count + trk->lfo_speed) & 0xff;
			detune += SIN(trk->lfo_count)
				* trk->lfo_power / 16 * trk->lfo_depth / 128;
			detune_switch = 1;
		}

		for (j = 0; j < trk->voicenum; j++) {
			if (trk->voice[j].porta_detune != 0) {
				s16 dtemp = trk->voice[j].porta_detune;
				trk->voice[j].porta_detune -= trk->voice[j].porta_step;
				if (ABS(trk->voice[j].porta_detune) > ABS(dtemp))
					trk->voice[j].porta_detune = 0;

				PDRV_KeySet(trk->voice[j].ch,
							trk->voice[j].note,
							trk->pitch + detune + trk->voice[j].porta_detune);
			} else {
				if (detune_switch)
					PDRV_KeySet(trk->voice[j].ch,
								trk->voice[j].note,
								trk->pitch + detune);
			}
		}
	}

	if (nmidi.count) {
		nmidi.count--;
		return;
	}

	do {
		u8 ch = pos->ch;
		u8 lsb = pos->lsb;
		u8 msb = pos->msb;
		NmidTrackStatus *trk = &nmidi.status[ch];

		switch (pos->mes & 0xf0) {
		case MES_NOTE_OFF: // 0x80: Note Off
			for (i = 0; i < trk->voicenum; i++) {
				if (trk->voice[i].note == lsb) {
// strict
//					if (pdrvw.stat[trk->voice[i].ch].note == lsb
//						&& pdrvw.stat[trk->voice[i].ch].env_phase > ENV_MOD_RR)
						PDRV_KeyOff(trk->voice[i].ch);
					for (j = i; j < trk->voicenum - 1; j++) {
						trk->voice[j].note = trk->voice[j+1].note;
						trk->voice[j].vel = trk->voice[j+1].vel;
						trk->voice[j].ch = trk->voice[j+1].ch;
						trk->voice[j].porta_detune = trk->voice[j+1].porta_detune;
						trk->voice[j].porta_step = trk->voice[j+1].porta_step;
					}
					trk->voice[j].note = 0;
					trk->voice[j].vel = 0;
					trk->voice[j].ch = 0;
					trk->voice[j].porta_detune = 0;
					trk->voice[j].porta_step = 0;
					trk->voicenum--;
					break;
				}
			}
			break;

		case MES_NOTE_ON: // 0x90: Note On
			{
				u32 level;
				s16 porta_step = 0;
				s16 porta_detune = 0;


				// Calcurate portament
				if (trk->porta_switch) {
					s16 key = trk->last_note - lsb;
					porta_detune = key * TICK_PITCH;
					porta_step = SFIXDIV(porta_detune, trk->porta_speed);
				}

				if (trk->bit.mute) 
					level = 0;
				else
					level = trk->vol * msb / 128;

				if (trk->bit.monopoly == NMID_TRK_MONO) {
					trk->voice[0].note = lsb;
					trk->voice[0].vel = msb;
					trk->voice[0].ch = ch;
					trk->voice[0].porta_detune = porta_detune;
					trk->voice[0].porta_step = porta_step;
					trk->last_note = lsb;

					PDRV_KeySet(ch, lsb, trk->pitch + porta_detune);
					PDRV_Volume(ch, level);
					if (trk->voicenum == 0) {
						PDRV_KeyOn(ch);
					}
					trk->voicenum = 1;
				} else
				if (trk->voicenum >= trk->voicemax) {
					u8 c = trk->voice[0].ch;
					if (trk->voicemax == 0)
						break;
					if (trk->voicemax == 1) {
						trk->voice[0].note = lsb;
						trk->voice[0].vel = msb;
						trk->voice[0].ch = c;
						trk->voice[0].porta_detune = porta_detune;
						trk->voice[0].porta_step = porta_step;
					} else {
						PDRV_KeyOff(trk->voice[0].ch);
						for (j = 0; j < trk->voicenum - 1; j++) {
//							if (trk->voice[j+1].note == 0x00) break;
							trk->voice[j].note = trk->voice[j+1].note;
							trk->voice[j].vel = trk->voice[j+1].vel;
							trk->voice[j].ch = trk->voice[j+1].ch;
							trk->voice[j].porta_detune = trk->voice[j+1].porta_detune;
							trk->voice[j].porta_step = trk->voice[j+1].porta_step;
						}
						trk->voice[j].note = lsb;
						trk->voice[j].vel = msb;
						trk->voice[j].ch = c;
						trk->voice[j].porta_detune = porta_detune;
						trk->voice[j].porta_step = porta_step;
					}
					trk->last_note = lsb;

					PDRV_KeySet(c, lsb, trk->pitch + porta_detune);
					PDRV_Volume(c, level);
					PDRV_KeyOn(c);
					break;
				} else {
					for (i = 0; i < NMID_MAX_CHANNELS; i++) {
						u32 mask = ((nmidi.channel_mask >> i) & 1);
						if (mask) continue;

						if (pdrvw.stat[i].env_phase == ENV_MOD_OFF ||
							pdrvw.stat[i].env_phase == ENV_MOD_RR) {
							trk->voice[trk->voicenum].note = lsb;
							trk->voice[trk->voicenum].vel = msb;
							trk->voice[trk->voicenum].ch = i;
							trk->voice[trk->voicenum].porta_detune = porta_detune;
							trk->voice[trk->voicenum].porta_step = porta_step;
							trk->voicenum++;
							trk->last_note = lsb;

							PCM_ProgramChange(i, trk->prog);
							PDRV_KeySet(i, lsb, trk->pitch + porta_detune);
							PDRV_KeyOn(i);
							PDRV_Volume(i, level);
							break;
						}
					}
				}
			}
			break;

		case MES_POLY_PRESSURE: // 0xA0: Polyphonic Key Pressure (not supported)
			break;

		case MES_CONTROL_CHANGE: // 0xB0: Control Change
			ControlChange(pos);
			break;

		case MES_PROGRAM_CHANGE: // 0xC0: Program Change
			trk->prog = lsb;
			PCM_ProgramChange(ch, lsb);
			break;

		case MES_CHANNEL_PRESSURE: // 0xD0: Channel Pressure (not Supported)
			break;

		case MES_PITCHBEND_CHANGE: // 0xE0: PitchBend Change
			trk->pitch = *((s16 *)&pos->lsb);
			for (i = 0; i < trk->voicenum; i++) {
				if (trk->voice[i].note != 0x00) {
					PDRV_KeySet(trk->voice[i].ch, 
					            trk->voice[i].note, 
					            trk->pitch);
				}
			}
			break;

		case MES_SYSTEM_MESSAGE: // 0xf0: System Message
			switch (ch) {
			case NMID_SYS_BPM:
				nmidi.bpm = lsb;
				break;
			case NMID_SYS_WAIT:
				nmidi.count = *((u16 *)&pos->lsb);
				break;
			case NMID_SYS_ENDDATA:
				if (nmidi.pos_loop != NULL) {
					nmidi.count = 0;
					pos = nmidi.pos_loop;
				} else {
					nmidi.count++;
//					pos--;
//					pos = nmidi.pos_start;
				}
				pos--;
				break;
			default:
				break;
			}
			break;
		default:
			break;
		}

		pos++;

	} while (!nmidi.count);

	nmidi.count--;
	nmidi.pos = pos;

#ifdef AGC_ENABLE
	// Auto Gain Control
	if (nmidi.agc) {
		u32 master_vol = 0;
		for (i = 0; i < MAX_CHANNELS; i++) {
			u8 phase = pdrvw.stat[i].env_phase;
			u8 v = pdrvw.stat[i].vol;
			u8 e = pdrvw.stat[i].env_vol;
			if (v && phase != ENV_MOD_OFF) {
				master_vol += v * e;
			}
		}
		if (master_vol > TICK_VOLUME * TICK_VOLUME) {
			master_vol /= TICK_VOLUME;
			master_vol = FIXDIV(TICK_VOLUME * TICK_VOLUME, master_vol);
		} else {
			master_vol = 128;
		}
		if (pdrvw.master_vol >= master_vol) {
			pdrvw.master_vol = master_vol;
			nmidi.agc_count = 0;
		} else {
			if (++nmidi.agc_count > nmidi.agc_release) {
				pdrvw.master_vol++;
				nmidi.agc_count = 0;
			}
		}
	}
#endif


}
