#ifndef NMID_H
#define NMID_H
#include <gba.h>

// #define AGC_ENABLE

#define NMID_MAX_CHANNELS	(16)
#define NMID_MAX_TRACKS		(16)

typedef struct nmid_dat {
	u8 mes;
	u8 ch;
	u8 lsb;
	u8 msb;
} ATTR_PACKED NMID_DAT;

typedef struct _NmidTrackStatus {
	u8 prog;			// Program Number (0~127)
	u8 vol;				// Volume (0-128)
	u8 exp;				// Expression (0-128)
	s8 pan;				// Pan (-128~127)

	s16 pitch;			// Pitchbend (-32768~32767)
	u8 porta_speed;		// Portament Speed (0-127)
	u8 porta_switch;	// Portament Switch

	u8 lfo_power;		// LFO Power
	u8 lfo_depth;		// LFO Depth
	u8 lfo_speed;		// LFO Speed
	u8 lfo_count;		// LFO Counter

	u8 voicemax;		// Max Voices
	u8 voicenum;		// Now Voices
	u8 last_note;		// Last Note (for portament)
	struct {
		u8 mode		:3;		//
		u8 mute		:1;		// 
		u8 monopoly	:1;		// 
	} bit;
	struct {
		u8 ch;			// channel (0-15)
		u8 note;		// Note Number (0~80)
		u8 vel;			// Velocity (0-128)
		u8 temp;
		s16 porta_step;		// Portament Step
		s16 porta_detune;	// Portament Detune
	} voice[NMID_MAX_CHANNELS];
} ATTR_PACKED NmidTrackStatus;

typedef struct nmid {
	NMID_DAT *pos;
	NMID_DAT *pos_start;
	NMID_DAT *pos_loop;
	u32 count;
	u32 channel_mask;
	u16 bpm;
	u16 temp;
#ifdef AGC_ENABLE
	u8 agc;
	u8 agc_release;
	u16 agc_count;
#endif
	NmidTrackStatus status[NMID_MAX_TRACKS];
} ATTR_PACKED NMID;

extern NMID nmidi IN_EWRAM;

void NMID_Init(void);
void NMID_SongSet(void *pos);
void NMID_Play(void);
void NMID_Mute(u8 ch, u8 flag);

#endif // NMID_H
