#ifndef _PDRV_H
#define _PDRV_H
#include <gba.h>
#include "pdrv_inst.h"

#define MAX_CHANNELS	(16)

/* */
#define MODE_PCM	0x00
#define MODE_PSG1	0x01
#define MODE_PSG2	0x02
#define MODE_PSG3	0x03
#define MODE_PSG4	0x04

/* envelope phase */
#define ENV_MOD_OFF	0x00
#define ENV_MOD_RR	0x01
#define ENV_MOD_SR	0x02
#define ENV_MOD_DR	0x04
#define ENV_MOD_AR	0x08

#define TICK_PITCH	(0x100)
#define TICK_RATE	(0x1000)
#define TICK_COUNT	((TICK_PITCH * TICK_RATE) / 16)
#define TICK_ENV	(0x0100)
#define TICK_DETUNE	(0x0100)
#define TICK_VTIMER	(0x1000)
#define TICK_VOLUME	(128)

typedef struct {
	INSTBANK *inst;	// Instrument Bank
	SPLIT *split;	// Split Bank
	u32 pcm_pos;	// PCM Position
	u32 pcm_cnt;	// PCM Counter
	u32 pcm_step;	// PCM Speed

	u8 mode;		// Mode (PCM or PSG)
	u8 note;		// Note Number
	s16 pitch;		// Pitchbend

	u8 env_rate;	// Envelope Rate
	u8 env_sl;		// Sustain Level
	u8 env_phase;	// Envelope Phase
	u8 env_vol;		// Calculated Volume

	u8 env_relvol;	// Temporary volume for release
	u8 env_cnt;		// Envelope Counter
	u8 vol;			// Volume
	s8 pan;			// pan (center = 0)
}  ATTR_PACKED PDRV_CHANNEL_STATUS;

typedef struct {
	u8 rateNum;
	u8 old_rateNum;
	u8 master_vol;
	s8 master_keyshift;
	u16 *rateTable;
	INSTBANK *inst;
	PDRV_CHANNEL_STATUS stat[MAX_CHANNELS];
}  ATTR_PACKED PDRVWORK;

void PDRV_Init(u16 rateNum, INSTBANK *);			// TvO[gԍwEFoÑAhXʒm
CODE_IN_IWRAM void PDRV_SetInstBank(INSTBANK *);	// 
CODE_IN_IWRAM u16 PDRV_SetRate(u16 rateNum);		// 
CODE_IN_IWRAM u16 PDRV_GetRate(void);				// TvO[gԂ
CODE_IN_IWRAM u16 PDRV_GetBuf(void);				// t[ɕKvȃTvobt@TCYԂ

CODE_IN_IWRAM void PCM_ProgramChange(int ch, u8 program);
CODE_IN_IWRAM void PDRV_KeySet(int ch, u8 note, s16 pitch);
CODE_IN_IWRAM void PDRV_KeyOn(int ch);
CODE_IN_IWRAM void PDRV_KeyOff(int ch);
CODE_IN_IWRAM void PDRV_Volume(int ch, u8 volume);
CODE_IN_IWRAM void PDRV_MasterVolume(u8 volume);

CODE_IN_IWRAM void PDRV_Render(u16 samples, s8 *buf);
CODE_IN_IWRAM void PDRV_Switch(s8 *buf);
#if 0
CODE_IN_IWRAM void PDRV_Render2(u16 samples, u8 *lbuf, u8 *rbuf);
CODE_IN_IWRAM void PDRV_Switch2(s8 *lbuf, s8 *rbuf);
#endif

#endif // _PDRV_H
