//==============================================================
// PDRV : PSG & PCM Module Driver
// 2005/01/22 Taro Yamada
//
// Description:
// 
//==============================================================
#include <gba.h>
#include <isr.h>
#include <gamemath.h>
#include "pdrv.h"

//==============================================================
// Prototypes
//==============================================================
CODE_IN_IWRAM static void PCM_KeySet(int ch, u8 note, s16 pitch);
CODE_IN_IWRAM static void PSG_KeySet(int ch, u8 note, s16 pitch);

//==============================================================
// Defines
//==============================================================
// #define PDRV_16KHZ
#define BASE_NOTE_NUMBER	(96)
#define ATTR_ALIGN4 __attribute__((aligned(4)))
#define ATTR_ALIGN2 __attribute__((aligned(2)))

#ifndef PDRV_16KHZ
#define RATENUM_MAX			(10)
#else
#define RATENUM_MAX			(12)
#endif

// mixing macro (1st)
// ~LVOobt@TCY K 16̔{Ȃ̂ŁA16JԂĂOK
#define mix_1st() \
{ \
				*mxr++ = samp; \
				cnt += step; \
				if (cnt > TICK_COUNT) { \
					pos += cnt / TICK_COUNT; \
					cnt &= (TICK_COUNT - 1); \
					if (pos > size) { \
						if (size > loop) { \
							do { \
								pos -= size; \
								pos += loop; \
							} while (pos > size); \
						} else { \
							stat->env_phase = ENV_MOD_OFF; \
							break; \
						} \
					} \
					samp = (snd_dat[pos] * vol); \
				} \
}
#define mix4_1st() { mix_1st(); mix_1st(); mix_1st(); mix_1st();}
#define mix16_1st() { mix4_1st(); mix4_1st(); mix4_1st(); mix4_1st(); }

// mixing macro (2nd-)
#define mix_add() \
{ \
				*mxr++ += samp; \
				cnt += step; \
				if (cnt > TICK_COUNT) { \
					pos += cnt / TICK_COUNT; \
					cnt &= (TICK_COUNT - 1); \
					if (pos > size) { \
						if (size > loop) { \
							do { \
								pos -= size; \
								pos += loop; \
							} while (pos > size); \
						} else { \
							stat->env_phase = ENV_MOD_OFF; \
							break; \
						} \
					} \
					samp = (snd_dat[pos] * vol); \
				} \
}
#define mix4_add() {mix_add(); mix_add(); mix_add(); mix_add();}
#define mix16_add() { mix4_add(); mix4_add(); mix4_add(); mix4_add(); }

//==============================================================
// Globals
//==============================================================
PDRVWORK pdrvw;
static s32 mixbuf[704];

//==============================================================
// Constant Data
//==============================================================
#ifndef PDRV_16KHZ
static const struct {
	u16 freq;
	u16 buf;
	u16 count;
	u16 cycle;
} RateTable[RATENUM_MAX] = {
	// freq  buf  count cycle (count+cycle=65536)
	{  5734,  96, 62610, 2926 }, // 0 :  5734 Hz _ 0x0060
	{ 10512, 176, 63940, 1596 }, // 1 : 10512 Hz _ 0x00B0
	{ 13378, 224, 64282, 1254 }, // 2 : 13378 Hz _ 0x00E0
	{ 18158, 304, 64612,  924 }, // 3 : 18158 Hz _ 0x0130
	{ 21024, 352, 64738,  798 }, // 4 : 21024 Hz _ 0x0160
	{ 26758, 448, 64909,  627 }, // 5 : 26758 Hz _ 0x01C0
	{ 31536, 528, 65004,  532 }, // 6 : 31536 Hz _ 0x0210
	{ 36314, 608, 65074,  462 }, // 7 : 36314 Hz _ 0x0260
	{ 40136, 672, 65118,  418 }, // 8 : 40136 Hz _ 0x02A0
	{ 42048, 704, 65137,  399 }, // 9 : 42048 Hz _ 0x02C0
};
#else
static const struct {
	u16 freq;
	u16 buf;
	u16 count;
	u16 cycle;
} RateTable[RATENUM_MAX] = {
	{  5734,  96, 62610, 2926 }, // 0 :  5734 Hz
	{  7884, 132, 63408, 2128 }, // 1 :  7884 Hz
	{ 10512, 176, 63940, 1596 }, // 2 : 10512 Hz
	{ 13378, 224, 64282, 1254 }, // 3 : 13378 Hz
	{ 15768, 264, 64472, 1064 }, // 4 : 15768 Hz
	{ 18158, 304, 64612,  924 }, // 5 : 18158 Hz
	{ 21024, 352, 64738,  798 }, // 6 : 21024 Hz
	{ 26758, 448, 64909,  627 }, // 7 : 26758 Hz
	{ 31536, 528, 65004,  532 }, // 8 : 31536 Hz
	{ 36314, 608, 65074,  462 }, // 9 : 36314 Hz
	{ 40136, 672, 65118,  418 }, // A : 40136 Hz
	{ 42048, 704, 65137,  399 }, // B : 42048 Hz
};
#endif
//static const u16 source_rate[] = {
//	11025, 16000, 22050, 32000,
//};

#ifndef PDRV_16KHZ
static const u16 rate_cnv[RATENUM_MAX][4] = {
//	   11025   16000   22050   32000 Hz
	{ 0x1ec3, 0x2ca5, 0x3d87, 0x594a, }, // 0 :  5734 Hz
	{ 0x10c7, 0x185a, 0x218f, 0x30b4, }, // 1 : 10512 Hz
	{ 0x0d2f, 0x1322, 0x1a5f, 0x2645, }, // 2 : 13378 Hz
	{ 0x09b6, 0x0e19, 0x136d, 0x1c32, }, // 3 : 18158 Hz
	{ 0x0863, 0x0c2d, 0x10c7, 0x185a, }, // 4 : 21024 Hz
	{ 0x0697, 0x0991, 0x0d2f, 0x1322, }, // 5 : 26758 Hz
	{ 0x0597, 0x081e, 0x0b2f, 0x103c, }, // 6 : 31536 Hz
	{ 0x04db, 0x070c, 0x09b7, 0x0e19, }, // 7 : 36314 Hz
	{ 0x0465, 0x0660, 0x08ca, 0x0cc1, }, // 8 : 40136 Hz
	{ 0x0431, 0x0616, 0x0863, 0x0c2d, }, // 9 : 42048 Hz
};
#else
static const u16 rate_cnv[RATENUM_MAX][4] = {
//	   11025   16000   22050   32000 Hz
	{ 0x1ec3, 0x2ca5, 0x3d87, 0x594a,  }, //  0 :  5734
	{ 0x165f, 0x2078, 0x2cbf, 0x40f1,  }, //  1 :  7884
	{ 0x10c7, 0x185a, 0x218f, 0x30b4,  }, //  2 : 10512
	{ 0x0d2f, 0x1322, 0x1a5f, 0x2645,  }, //  3 : 13378
	{ 0x0b2f, 0x103c, 0x165f, 0x2078,  }, //  4 : 15768
	{ 0x09b6, 0x0e19, 0x136d, 0x1c32,  }, //  5 : 18158
	{ 0x0863, 0x0c2d, 0x10c7, 0x185a,  }, //  6 : 21024
	{ 0x0697, 0x0991, 0x0d2f, 0x1322,  }, //  7 : 26758
	{ 0x0597, 0x081e, 0x0b2f, 0x103c,  }, //  8 : 31536
	{ 0x04db, 0x070c, 0x09b7, 0x0e19,  }, //  9 : 36314
	{ 0x0465, 0x0660, 0x08ca, 0x0cc1,  }, //  A : 40136
	{ 0x0431, 0x0616, 0x0863, 0x0c2d,  }, //  B : 42048
};
#endif
static const u16 gbs_fnum[] = {
//       C      C#       D      D#       E       F      F#       G      G#       A      A#       B
	     0,      0,      0,      0,      0,      0,      0,      0,      0,      0,      0,      0,  //  0-11
	     0,      0,      0,      0,      0,      0,      0,      0,      0,      0,      0,      0,  // 12-23
	0x002c, 0x009d, 0x0107, 0x016b, 0x01c9, 0x0223, 0x0277, 0x02c7, 0x0312, 0x0359, 0x039c, 0x03db,  // 24-35
	0x0416, 0x044e, 0x0484, 0x04b5, 0x04e5, 0x0511, 0x0534, 0x0563, 0x0589, 0x05ac, 0x05c9, 0x05e9,  // 36-47
	0x060b, 0x0627, 0x063f, 0x0658, 0x0670, 0x0687, 0x069c, 0x06b0, 0x06c3, 0x06d5, 0x06e6, 0x06f6,  // 48-59
	0x0705, 0x0713, 0x0721, 0x072d, 0x0739, 0x0744, 0x074f, 0x0759, 0x0762, 0x076b, 0x0773, 0x077b,  // 60-71
	0x0783, 0x078a, 0x0790, 0x0797, 0x079d, 0x07a2, 0x07a7, 0x07ac, 0x07b1, 0x07b6, 0x07ba, 0x07be,  // 72-13
	0x07c1, 0x07c5, 0x07c8, 0x07cb, 0x07ce, 0x07d1, 0x07d4, 0x07d6, 0x07d9, 0x07db, 0x07dd, 0x07df,  // 84-95
	0x07e1, 0x07e2, 0x07e4, 0x07e6, 0x07e7, 0x07e9, 0x07ea, 0x07eb, 0x07ec, 0x07ed, 0x07ee, 0x07ef,  // 96-107
	0x07f0, 0x07f1, 0x07f2, 0x07f3, 0x07f4, 0x07f5, 0x07f6, 0x07f7, 0x07f8, 0x07f9, 0x07fa, 0x07fb,  // 108-119
	0x07fc, 0x07fd, 0x07fe, 0x07ff, 0x07ff, 0x07ff, 0x07ff, 0x07ff,                                  // 120-127
};
static const u16 scc_fnum[] = {
//       C      C#       D      D#       E       F      F#       G      G#       A      A#       B
	     0,      0,      0,      0,      0,      0,      0,      0,      0,      0,      0,      0,  //  0-11
	0x0029, 0x009d, 0x0107, 0x016b, 0x01c9, 0x0223, 0x0277, 0x02c7, 0x0312, 0x0359, 0x039b, 0x03da,  // 12-23
	0x0416, 0x044e, 0x0483, 0x04b5, 0x04e5, 0x0512, 0x053b, 0x0563, 0x0589, 0x05ad, 0x05cd, 0x05ed,  // 24-35
	0x060b, 0x0627, 0x0642, 0x065b, 0x0672, 0x0689, 0x069e, 0x06b2, 0x06c4, 0x06d6, 0x06e7, 0x06f7,  // 36-47
	0x0706, 0x0714, 0x0721, 0x072d, 0x0739, 0x0744, 0x074f, 0x0759, 0x0762, 0x076b, 0x0773, 0x077b,  // 48-59
	0x0783, 0x078a, 0x0790, 0x0796, 0x079c, 0x07a2, 0x07a7, 0x07ac, 0x07b1, 0x07b6, 0x07ba, 0x07be,  // 60-71
	0x07c1, 0x07c5, 0x07c8, 0x07cb, 0x07ce, 0x07d1, 0x07d4, 0x07d6, 0x07d9, 0x07db, 0x07dd, 0x07df,  // 72-13
	0x07e1, 0x07e2, 0x07e4, 0x07e6, 0x07e7, 0x07e9, 0x07ea, 0x07eb, 0x07ec, 0x07ed, 0x07ee, 0x07ef,  // 84-95
	0x07f0, 0x07f1, 0x07f2, 0x07f3, 0x07f4, 0x07f5, 0x07f6, 0x07f7, 0x07f8, 0x07f9, 0x07fa, 0x07fb,  // 96-107
	0x07fc, 0x07fd, 0x07fe, 0x07ff, 0x07ff, 0x07ff, 0x07ff, 0x07ff, 0x07ff, 0x07ff, 0x07ff, 0x07ff,  // 108-119
	0x07ff, 0x07ff, 0x07ff, 0x07ff, 0x07ff, 0x07ff, 0x07ff, 0x07ff,                                  // 120-127
};

static const u16 pcm_fnum[] = {
//       C      C#       D      D#       E       F      F#       G      G#       A      A#       B       C
	0x1000, 0x10f4, 0x11f6, 0x1307, 0x1428, 0x155b, 0x16a1, 0x17f9, 0x1966, 0x1ae9, 0x1c83, 0x1e35, 0x2000,
};

static const struct {
	s8	oct;
	u8	key;
} ATTR_PACKED Shiftable[] = {
//      C      C#       D      D#       E       F      F#       G      G#       A       A#        B
	{-8,0}, {-8,1}, {-8,2}, {-8,3}, {-8,4}, {-8,5}, {-8,6}, {-8,7}, {-8,8}, {-8,9}, {-8,10}, {-8,11}, // o-4    0 -  11
	{-7,0}, {-7,1}, {-7,2}, {-7,3}, {-7,4}, {-7,5}, {-7,6}, {-7,7}, {-7,8}, {-7,9}, {-7,10}, {-7,11}, // o-3   12 -  23
	{-6,0}, {-6,1}, {-6,2}, {-6,3}, {-6,4}, {-6,5}, {-6,6}, {-6,7}, {-6,8}, {-6,9}, {-6,10}, {-6,11}, // o-2   24 -  35
	{-5,0}, {-5,1}, {-5,2}, {-5,3}, {-5,4}, {-5,5}, {-5,6}, {-5,7}, {-5,8}, {-5,9}, {-5,10}, {-5,11}, // o-1   36 -  47
	{-4,0}, {-4,1}, {-4,2}, {-4,3}, {-4,4}, {-4,5}, {-4,6}, {-4,7}, {-4,8}, {-4,9}, {-4,10}, {-4,11}, // o0    48 -  59
	{-3,0}, {-3,1}, {-3,2}, {-3,3}, {-3,4}, {-3,5}, {-3,6}, {-3,7}, {-3,8}, {-3,9}, {-3,10}, {-3,11}, // o1    60 -  71
	{-2,0}, {-2,1}, {-2,2}, {-2,3}, {-2,4}, {-2,5}, {-2,6}, {-2,7}, {-2,8}, {-2,9}, {-2,10}, {-2,11}, // o2    72 -  13
	{-1,0}, {-1,1}, {-1,2}, {-1,3}, {-1,4}, {-1,5}, {-1,6}, {-1,7}, {-1,8}, {-1,9}, {-1,10}, {-1,11}, // o3    84 -  95

	{ 0,0}, { 0,1}, { 0,2}, { 0,3}, { 0,4}, { 0,5}, { 0,6}, { 0,7}, { 0,8}, { 0,9}, { 0,10}, { 0,11}, // o4    96 - 107

	{ 1,0}, { 1,1}, { 1,2}, { 1,3}, { 1,4}, { 1,5}, { 1,6}, { 1,7}, { 1,8}, { 1,9}, { 1,10}, { 1,11}, // o5   108 - 119
	{ 2,0}, { 2,1}, { 2,2}, { 2,3}, { 2,4}, { 2,5}, { 2,6}, { 2,7}, { 2,8}, { 2,9}, { 2,10}, { 2,11}, // o6   120 - 127
	{ 3,0}, { 3,1}, { 3,2}, { 3,3}, { 3,4}, { 3,5}, { 3,6}, { 3,7}, { 3,8}, { 3,9}, { 3,10}, { 3,11}, // o7   128 - 139
	{ 4,0}, { 4,1}, { 4,2}, { 4,3}, { 4,4}, { 4,5}, { 4,6}, { 4,7}, { 4,8}, { 4,9}, { 4,10}, { 4,11}, // o8   140 - 151
	{ 5,0}, { 5,1}, { 5,2}, { 5,3}, { 5,4}, { 5,5}, { 5,6}, { 5,7}, { 5,8}, { 5,9}, { 5,10}, { 5,11}, // o9   152 - 163
	{ 6,0}, { 6,1}, { 6,2}, { 6,3}, { 6,4}, { 6,5}, { 6,6}, { 6,7}, { 6,8}, { 6,9}, { 6,10}, { 6,11}, // o10  164 - 175
	{ 7,0}, { 7,1}, { 7,2}, { 7,3}, { 7,4}, { 7,5}, { 7,6}, { 7,7}, { 7,8}, { 7,9}, { 7,10}, { 7,11}, // o11  176 - 187
	{ 8,0}, { 8,1}, { 8,2}, { 8,3}, { 8,4}, { 8,5}, { 8,6}, { 8,7}, { 8,8}, { 8,9}, { 8,10}, { 8,11}, // o12  188 - 199
};

//==============================================================
// Init PDRV
//==============================================================
void PDRV_Init(u16 rateNum, INSTBANK *inst)
{
	int i;

	if (rateNum >= RATENUM_MAX) {
		rateNum = RATENUM_MAX - 1;
	}
	pdrvw.rateNum = rateNum;
	pdrvw.rateTable = (u16 *)&rate_cnv[rateNum][0];
	pdrvw.inst = inst;
	pdrvw.master_vol = 128;
	pdrvw.master_keyshift = 0;
	
	for (i = 0; i < MAX_CHANNELS; i++) {
		pdrvw.stat[i].inst = &pdrvw.inst[0];
		pdrvw.stat[i].split = &pdrvw.inst[0].bank[0];
		
		pdrvw.stat[i].pcm_pos = 0;	// PCM Position
		pdrvw.stat[i].pcm_cnt = 0;	// PCM Counter
		pdrvw.stat[i].pcm_step = 0;	// PCM Speed

		pdrvw.stat[i].mode = MODE_PCM;	// MODE = PCM
		pdrvw.stat[i].note = 0;		// Note Number
		pdrvw.stat[i].pitch = 0;	// Pitchbend

		pdrvw.stat[i].env_rate = 0;
		pdrvw.stat[i].env_sl = 0;
		pdrvw.stat[i].env_phase = ENV_MOD_OFF;
		pdrvw.stat[i].env_vol = 128;
		pdrvw.stat[i].env_relvol = 128;

		pdrvw.stat[i].env_cnt = 0;
		pdrvw.stat[i].vol = 128;
		pdrvw.stat[i].pan = 0;
	}

	REG_TM0D = RateTable[pdrvw.rateNum].count;
	REG_TM0CNT = TIMER_PRESCALER_1CLK | TIMER_ENABLE;
	REG_DMA1DAD = &REG_SGFIFOA;
	REG_SGCNT1 = SG1_SOUND_1_ON
				| SG1_SOUND_2_ON
				| SG1_SOUND_3_ON
				| SG1_SOUND_4_ON
				| SG1_SOUND_ALL_ON;
	return;
}

//==============================================================
// Set/Get Generater Sampling Rate
//==============================================================
// FoÑAhXn
CODE_IN_IWRAM
void 
PDRV_SetInstBank(INSTBANK *inst)
{
	pdrvw.inst = inst;
	return;
}
// TvO[g(ԍ)n
// Ԃl͕Kvobt@TCY
CODE_IN_IWRAM
u16 
PDRV_SetRate(u16 rateNum)
{
	int i;

	if (rateNum >= RATENUM_MAX) {
		rateNum = RATENUM_MAX - 1;
	}

	if (rateNum == pdrvw.rateNum)
		return RateTable[rateNum].buf;

	pdrvw.rateNum = rateNum;
	pdrvw.rateTable = (u16 *)&rate_cnv[rateNum][0];

	// `lpcm_stepvZȂ
	for (i = 0; i < MAX_CHANNELS; i++) {
		PDRV_CHANNEL_STATUS *ch_state = &pdrvw.stat[i];
		if (ch_state->mode == MODE_PCM
			&& ch_state->env_phase != ENV_MOD_OFF) {
			PCM_KeySet(i, ch_state->note, ch_state->pitch);
		}
	}

	return RateTable[rateNum].buf;
}
// TvO[gԂ
CODE_IN_IWRAM
u16 
PDRV_GetRate(void)
{
	return RateTable[pdrvw.rateNum].freq;
}
// t[ɕKvȃTvobt@TCYԂ
CODE_IN_IWRAM
u16 
PDRV_GetBuf(void)
{
	return RateTable[pdrvw.rateNum].buf;
}

//==============================================================
// Program Change
//==============================================================
CODE_IN_IWRAM
void 
PCM_ProgramChange(int ch, u8 program)
{
	PDRV_CHANNEL_STATUS *stat = (PDRV_CHANNEL_STATUS *)&pdrvw.stat[ch];
	stat->inst = &pdrvw.inst[program];
	return;
}

CODE_IN_IWRAM
void 
PSG_ProgramChange(int ch, u8 program)
{
	PDRV_CHANNEL_STATUS *stat = (PDRV_CHANNEL_STATUS *)&pdrvw.stat[ch];
	stat->inst = &pdrvw.inst[program];
	return;
}

//==============================================================
// Pitch Change
//==============================================================
CODE_IN_IWRAM
static 
void 
PCM_KeySet(int ch, u8 note, s16 pitch)
{
	INSTBANK *inst = pdrvw.stat[ch].inst;
	u8 key_margin;
	u8 key;
	s8 oct;
	u16 temp_note;
	u32 step;
	int i;

	note += pdrvw.master_keyshift;

	// XvbgrāAĐ\ȃCXgDgT
	// iȂƂ̂Ƃ͍lĖ̂ŒӁj
	for (i = 0; i < MAX_SPLIT; i++) {
		if (note <= inst->bank[i].split) {
			if (i > 0 && inst->bank[i-1].split >= inst->bank[i].split)
				i = 7;
			break;
		}
	}

	// m[gio[Ƀsb`xh𔽉f
	temp_note = (note * TICK_PITCH) + pitch;
	note = (temp_note / TICK_PITCH);
	pitch = temp_note & (TICK_PITCH - 1);

	// CXgDg̃AhXۑ
	pdrvw.stat[ch].split = &inst->bank[i];
	// noterootkey̍特߂
	key_margin = BASE_NOTE_NUMBER + note - inst->bank[i].rootkey;

	// IN^[uVtglƃL[߂
	oct = Shiftable[key_margin].oct;
	key = Shiftable[key_margin].key;

	// Xebv߂
	// 16bit * 16bit / 8bit = 24bit
	if (pitch == 0) {
		step = (pcm_fnum[key] * pdrvw.rateTable[inst->bank[i].snd->freqNum]) / 256;
	} else {
		u16 pitch_margin = pcm_fnum[key+1] - pcm_fnum[key];
		pitch_margin = pitch_margin * pitch / TICK_PITCH;
		step = ((pcm_fnum[key] + pitch_margin) * pdrvw.rateTable[inst->bank[i].snd->freqNum]) / 256;
	}

	// ő8bitVtg
	if (oct < 0) {
		step >>= -oct;
	} else {
		step <<= oct;
	}

	pdrvw.stat[ch].pcm_step = step;
}

CODE_IN_IWRAM
static 
void 
PSG_KeySet(int ch, u8 note, s16 pitch)
{
	u16 *fnumber;
	u16 temp_note;
	u16 step;

	// m[gio[Ƀsb`xh𔽉f
	temp_note = (note * TICK_PITCH) + pitch;
	note = (temp_note / TICK_PITCH);
	pitch = temp_note & (TICK_PITCH - 1);

	switch (ch) {
	case 0-1:
		fnumber = (u16 *)&gbs_fnum[note];
		break;
	case 2:
		fnumber = (u16 *)&scc_fnum[note];
		break;
	case 3:
		fnumber = (u16 *)&gbs_fnum[note];
		break;
	default:
		break;
	}

	if (pitch == 0) {
		step = *fnumber;
	} else {
		u16 pitch_margin = *(fnumber+1) - *fnumber;
		pitch_margin = FIXDIV(pitch_margin * 256, pitch);
		step = (gbs_fnum[note] + pitch_margin);
	}
	
}

CODE_IN_IWRAM
void 
PDRV_KeySet(int ch, u8 note, s16 pitch)
{
	PDRV_CHANNEL_STATUS *pPdrvChannelState = &pdrvw.stat[ch];

	pPdrvChannelState->note = note;
	pPdrvChannelState->pitch = pitch;

	switch (pPdrvChannelState->mode) {
	case MODE_PCM:
		PCM_KeySet(ch, note, pitch);
		break;
	case MODE_PSG1:
		PSG_KeySet(ch, note, pitch);
		break;
	case MODE_PSG2:
		PSG_KeySet(ch, note, pitch);
		break;
	case MODE_PSG3:
		PSG_KeySet(ch, note, pitch);
		break;
	case MODE_PSG4:
		break;
	default:
		PCM_KeySet(ch, note, pitch);
		break;
	}
	return;
}

CODE_IN_IWRAM
void 
PDRV_ModeSet(int ch, int mode)
{
	
}

//==============================================================
// Key On/Off
//==============================================================
CODE_IN_IWRAM
void 
PDRV_KeyOn(int ch)
{
	PDRV_CHANNEL_STATUS *stat = (PDRV_CHANNEL_STATUS *)&pdrvw.stat[ch];
	stat->pcm_pos = 0;
	stat->pcm_cnt = 0;
	stat->env_cnt = 0;
	stat->env_vol = 0;
	stat->env_phase = ENV_MOD_AR;
	stat->env_rate = stat->split->ar;
	stat->env_sl = stat->split->sl;
	return;
}

CODE_IN_IWRAM
void 
PDRV_KeyOff(int ch)
{
	PDRV_CHANNEL_STATUS *stat = (PDRV_CHANNEL_STATUS *)&pdrvw.stat[ch];

	if (stat->split->rr == 0x00) {
		stat->env_vol = 0;
		stat->env_phase = ENV_MOD_OFF;
		return;
	}
	if (stat->env_phase != ENV_MOD_OFF) {
		stat->env_cnt = 0;
		stat->env_phase = ENV_MOD_RR;
		stat->env_rate = stat->split->rr;
		stat->env_relvol = stat->env_vol;
	}
	return;
}

//==============================================================
// Volume Set
//==============================================================
CODE_IN_IWRAM
void 
PDRV_Volume(int ch, u8 volume)
{
	PDRV_CHANNEL_STATUS *stat = (PDRV_CHANNEL_STATUS *)&pdrvw.stat[ch];
	if (volume > TICK_VOLUME)
		volume = TICK_VOLUME;
	stat->vol = volume;
	return;
}

CODE_IN_IWRAM
void 
PDRV_MasterVolume(u8 volume)
{
	pdrvw.master_vol = volume;
}

/*****************************************************************************
 Gx[v
 A^bN->fBPC->TXeB->[X
*****************************************************************************/
IN_IWRAM
static void 
PDRV_PhaseEnvelope(void)
{
	int ch = MAX_CHANNELS;
	PDRV_CHANNEL_STATUS *ENV = (PDRV_CHANNEL_STATUS *)&pdrvw.stat[0];
	while (ch > 0) {
		if (ENV->env_phase == ENV_MOD_OFF) {
			ENV++;
			ch--;
			continue;
		}

		switch (ENV->env_phase) {
		case ENV_MOD_AR:
			// A^bN Attack = Counter / AttackRate
			if (ENV->env_rate > ENV->env_cnt) {
				ENV->env_cnt++;
				ENV->env_vol = FIXDIV(TICK_VOLUME * ENV->env_cnt, ENV->env_rate);
				break;
			}
			ENV->env_phase = ENV_MOD_DR;
			ENV->env_rate = ENV->split->dr;
			ENV->env_vol = TICK_VOLUME;
			ENV->env_cnt = 0;
		case ENV_MOD_DR:
			// fBPC
			if (ENV->env_rate == 255) {
				break;
			}
			if (ENV->env_rate > ENV->env_cnt) {
				ENV->env_cnt++;
				ENV->env_vol = TICK_VOLUME - FIXDIV((TICK_VOLUME - ENV->env_sl) * ENV->env_cnt, ENV->env_rate);
				break;
			}
			ENV->env_phase = ENV_MOD_SR;
			ENV->env_rate = ENV->split->sr;
			ENV->env_cnt = 0;
		case ENV_MOD_SR:
			// TXeB
			if (ENV->env_rate == 255) {
				ENV->env_vol = ENV->env_sl;
				break;
			}
			if (ENV->env_rate > ENV->env_cnt) {
				ENV->env_cnt++;
				ENV->env_vol = ENV->env_sl - FIXDIV(ENV->env_sl * ENV->env_cnt, ENV->env_rate);
				break;
			}
			ENV->env_phase = ENV_MOD_OFF;
			ENV->env_cnt = 0;
			ENV->env_vol = 0;
			ENV->pcm_pos = 0;
			ENV->pcm_cnt = 0;
			break;
		case ENV_MOD_RR:
			// [X
			if (ENV->env_rate == 255) {
				// RR = 255 ̂Ƃvolێ
				break;
			}
			ENV->env_cnt++;
			if (ENV->env_rate > ENV->env_cnt) {
				ENV->env_vol = ENV->env_relvol - FIXDIV(ENV->env_relvol * ENV->env_cnt, ENV->env_rate);
				break;
			}
			ENV->env_phase = ENV_MOD_OFF;
			ENV->env_cnt = 0;
			ENV->env_vol = 0;
			ENV->pcm_pos = 0;
			break;
		default:
			ENV->env_phase = ENV_MOD_OFF;
			ENV->env_cnt = 0;
			ENV->env_vol = 0;
			ENV->pcm_pos = 0;
			break;
		}
		ENV++;
		ch--;
	}
	return;
}



//==============================================================
// Render Audio
//==============================================================
CODE_IN_IWRAM
void 
PDRV_Render(u16 samples, s8 *buf)
{
	int i = 0;
	int first_render = 1;

	PDRV_PhaseEnvelope();

	for (i = 0; i < MAX_CHANNELS; i++) {
		PDRV_CHANNEL_STATUS *stat = (PDRV_CHANNEL_STATUS *)&pdrvw.stat[i];
		SND *snd = stat->split->snd;
		s8 *snd_dat = &snd->data[0];
		u32 length = samples;

		register s32 samp;
		s32 *mxr = mixbuf;
		u32 pos = stat->pcm_pos;
		u32 cnt = stat->pcm_cnt;
		const u32 step = stat->pcm_step;
		const u32 size = snd->size;
		const u32 loop = snd->loop;
		const u32 vol = stat->env_vol * stat->vol * pdrvw.master_vol / (TICK_VOLUME * TICK_VOLUME);
		// ~Ȃ珈΂
		if (stat->env_phase == ENV_MOD_OFF)
			continue;
		// {[łقǏ΁A~LVO΂
		if (vol < 1) {
			u32 c = step * (samples / 128);
			pos += c / (TICK_COUNT / 128);
			if (pos > size) {
				if (size > loop) {
					do {
						pos -= size;
						pos += loop;
					} while (pos > size);
				} else {
					stat->env_phase = ENV_MOD_OFF;
					break;
				}
			}
			stat->pcm_pos = pos;
			stat->pcm_cnt = cnt;
			continue;
		}
		samp = (snd_dat[pos] * vol);
		if (first_render) {
			length /= 16;
			do {
				mix16_1st();
			} while (--length);
			first_render = 0;
		} else {
			length /= 16;
			do {
				mix16_add();
			} while (--length);
		}

		stat->pcm_pos = pos;
		stat->pcm_cnt = cnt;
	}

	// NbsO
	if (first_render) {
		u32 length = samples / 16;
		u32 *mxr = (u32 *)buf;
		do {
			*mxr++ = 0;
			*mxr++ = 0;
			*mxr++ = 0;
			*mxr++ = 0;
		} while (--length);
	} else {
		u32 length = samples;
		register s8 *mxr = buf;
		register s32 *mx = mixbuf;
		do {
			register s32 m = *mx++;
			m /= TICK_VOLUME;
			if (m > 127) {
				m = 127;
			} else if (m < -127) {
				m = -127;
			}
			*mxr++ = (s8)m;
		} while (--length);
	}
}

#if 0
CODE_IN_IWRAM
void 
PDRV_Render2(u16 samples, u8 *lbuf, u8 *rbuf)
{
	int i, j;
	for (i = 0; i < MAX_CHANNELS; i++) {
		PDRV_CHANNEL_STATUS *stat = (PDRV_CHANNEL_STATUS *)&pdrvw.stat[i];
		SND *snd = stat->split->snd;
		s8 *snd_dat = &stat->split->snd->data[0];
		u8 left = stat->env_vol * stat->vol / TICK_VOLUME;
		u8 right = stat->env_vol * stat->vol / TICK_VOLUME;

		// ~A{[łقǏ΁A΂
		if (stat->env_phase == ENV_MOD_OFF || stat->env_vol <= 1)
			continue;

		for (j = 0; j < samples; j++) {
			lbuf[j] += (snd_dat[stat->pcm_pos] * left) / TICK_VOLUME;
			rbuf[j] += (snd_dat[stat->pcm_pos] * right) / TICK_VOLUME;
			stat->pcm_cnt += stat->pcm_step;
			while (stat->pcm_cnt > TICK_COUNT) {
				stat->pcm_pos++;
				stat->pcm_cnt -= TICK_COUNT;
				// I[܂ōĐꍇA[v|Cg΃[v
				if (stat->pcm_pos > snd->size) {
					if (snd->size > snd->loop) {
						stat->pcm_pos = snd->loop;
					} else {
						stat->env_phase = ENV_MOD_OFF;
						continue;
					}
				}
			}
		}
	}
}
#endif

//==============================================================
// Switch Audio Buffer
//==============================================================
CODE_IN_IWRAM
void 
PDRV_Switch(s8 *buf)
{
	REG_DMA1CNT = 0;
	REG_DMA1SAD = (u32 *)buf;
	REG_DMA1CNT = DMA_DEST_FIX | DMA_SRC_INC | DMA_REPEAT | DMA_16BIT | DMA_FIFO | DMA_ENABLE | 1;
	if (pdrvw.old_rateNum != pdrvw.rateNum) {
		REG_TM0CNT = 0;
		REG_TM0D = RateTable[pdrvw.rateNum].count;
		REG_TM0CNT = TIMER_PRESCALER_1CLK | TIMER_ENABLE;
		pdrvw.old_rateNum = pdrvw.rateNum;
	}
}

#if 0
// DMG_A = L, DMG_B = R
CODE_IN_IWRAM
void 
PDRV_Switch2(s8 *lbuf, s8 *rbuf)
{
/*	REG_SGCNT0_H = 
		SG0_SOUND_OUT_FULL
		| SG0_DSOUND_A_OUT_FULL
		| SG0_DSOUND_B_OUT_FULL
		| SG0_DSOUND_A_L_ON
		| SG0_DSOUND_A_TIMER0
		| SG0_DSOUND_A_RESET
		| SG0_DSOUND_B_R_ON
		| SG0_DSOUND_B_TIMER1
		| SG0_DSOUND_B_RESET;
*/
	REG_DMA1SAD = (u32 *)lbuf;
	REG_DMA2SAD = (u32 *)rbuf;
	REG_DMA1DAD = &REG_SGFIFOA;
	REG_DMA2DAD = &REG_SGFIFOB;
	REG_DMA1CNT = DMA_DEST_FIX | DMA_SRC_INC | DMA_32BIT | DMA_FIFO | DMA_ENABLE;
	REG_DMA2CNT = DMA_DEST_FIX | DMA_SRC_INC | DMA_32BIT | DMA_FIFO | DMA_ENABLE;
	REG_TM0D = TIMER_DISABLE;
	REG_TM1D = TIMER_DISABLE;
	REG_TM0CNT = RateTable[pdrvw.rateNum].count;
	REG_TM1CNT = RateTable[pdrvw.rateNum].count;
	REG_TM0D = TIMER_PRESCALER_1CLK | TIMER_ENABLE;
	REG_TM1D = TIMER_PRESCALER_1CLK | TIMER_ENABLE;
}
#endif
