my %mode = (
	'256' => 0,
	'pal' => 0,
);

&getarg();
open(FILE,"$src") || die "Can't open $src\n";
binmode FILE;
&read_fileheader();
&read_infoheader();
&read_palette();
print "image : $Width x $Height ($BitSize bit $ColSize color)\n";

&read_bitmap();
close(FILE);

if ($dst_c ne "") {
	open(FILE, ">$dst_c") || die "Can't open $dst_c\n";
	print FILE "#include \"$fn.h\"\n\n";
	if ($mode{'pal'}) {
		&write_palette();
	}
	&write_bitmap();
	close (FILE);
	print "$dst_c written\n";
}
if ($dst_h ne "") {
	open(FILE, ">$dst_h") || die "Can't open $dst_h\n";
	print FILE "#ifndef _$constatntname\_H\n";
	print FILE "#define _$constatntname\_H\n";
	print FILE "#include <gba.h>\n\n";
	if ($mode{'pal'}) {
		print FILE "extern const u16 $constatntname\_pal[];\n";
	}
	print FILE "extern const u16 $constatntname\_chr[];\n\n";
	print FILE "#endif\n";
	close (FILE);
	print "$dst_h written\n";
}

exit 0;

# ---------------------------------------------------------------------------
# 
sub getarg
{
	my $ARGC = @ARGV;
	my $tmp;
	my $optionnum;
	&prterror("\nUSAGE : b2s [/256] [/pal] src[.bmp] [dst]\n") if ($ARGV[0] eq '');
	$optionnum = &getoption(@ARGV);

	$src = $ARGV[$optionnum];
	$dst = $ARGV[$optionnum+1] if($ARGV[1] ne '');
	$tmp = $src =~ m/\.bmp$/i;
	if ($tmp) {
		$fn  = $src;
		$fn  =~ s/\.bmp$//i;
		$dst_h = $fn.".h" if($ARGV[$optionnum+1] eq '');
		$dst_c = $fn.".c" if($ARGV[$optionnum+1] eq '');
	} else {
		$fn = $src;
		$dst_h = $src.".h" if($ARGV[$optionnum+1] eq '');
		$dst_c = $src.".c" if($ARGV[$optionnum+1] eq '');
		$src .= ".bmp";
	}
	$fn =~ s/.*\///;
	$fn =~ /([^\/]*)$/;
	$constatntname = $1;
	return;
}
sub getoption
{
	my @tmp = @_[0];
	my $i;
	my $optionnum = 0;
	for ($i = 0; $i < @tmp; $i++) {
		if ($tmp[$i] =~ /^\//) {
			$tmp[$i] =~ m/^\/([0-9a-zA-Z]+)/;
			$mode{"$1"} = 1;
			$optionnum++;
		}
		if ($tmp[$i] =~ /^-/) {
			$tmp[$i] =~ m/^-([0-9a-zA-Z]+)/;
			$mode{"$1"} = 1;
			$optionnum++;
		}
	}
#	foreach (keys %mode) {
#		print "$_ : $mode{$_}\n";
#	}
	return $optionnum;
}

# ---------------------------------------------------------------------------
# 
sub read_fileheader
{
	my $tmp;
	read (FILE, $Magic, 2);
	&prterror("Illegal BMP header!\nAbort...\n") if ($Magic ne 'BM');
	read (FILE, $FileSize, 4);
	read (FILE, $tmp, 4);
	read (FILE, $ImageOffset, 4);
	$FileSize    = unpack("L", $FileSize);
	$ImageOffset = unpack("L", $ImageOffset);
}

# ---------------------------------------------------------------------------
# 
sub read_infoheader
{
	my $tmp;
	read (FILE, $InfoHeaderSize, 4);
	read (FILE, $Width, 4);
	read (FILE, $Height, 4);
	read (FILE, $Planes, 2);
	read (FILE, $BitSize, 2);
	read (FILE, $Compress, 4);
	read (FILE, $HDpi, 4);
	read (FILE, $VDpi, 4);
	read (FILE, $PalUsed, 4);
	read (FILE, $PalImportant, 4);

	$InfoHeaderSize = unpack("V", $InfoHeaderSize);
	$Width = unpack("V", $Width);
	$Height = unpack("V", $Height);
	$BitSize = unpack("v", $BitSize);
	$ColSize = 1 << $BitSize;
	$PalUsed = unpack("V", $PalUsed);
}

# ---------------------------------------------------------------------------
# 
sub read_palette
{
	my $r,$g,$b,$tmp;
	&prterror("8bit only, it uses $BitSize\n") if ($ColSize != 256);
	seek (FILE, 14+$InfoHeaderSize, 0);
	for ($i = 0; $i < $ColSize; $i++) {
		read (FILE, $b, 1);	$b = unpack("C", $b); $b = int ( $b / 8 );
		read (FILE, $g, 1);	$g = unpack("C", $g); $g = int ( $g / 8 );
		read (FILE, $r, 1);	$r = unpack("C", $r); $r = int ( $r / 8 );
		read (FILE, $tmp, 1); # reserved
		$tmp = ($b<<10) + ($g<<5) + ($r);
#		$tmp = pack("n*", $tmp);
#		$tmp = unpack("v*", $tmp);
		unshift(@palette, $tmp);
	}
}

# ---------------------------------------------------------------------------
# 
sub read_bitmap
{
	my $tmp;
	seek (FILE, $ImageOffset, 0);
	for ($i = 0; $i < $Height; $i++) {
		$bitmap[$Height -1 - $i] = "";
		for ($j = 0; $j < $Width; $j++) {
			read (FILE, $tmp, 1);
			$tmp = unpack("C", $tmp);
			unless ($mode{'256'}) {
				$tmp &= 0x0f;
				$tmp = sprintf("%01X",$tmp);
			} else {
				$tmp = sprintf("%02X",$tmp);
			}
			$bitmap[$Height -1 - $i] .= $tmp;
		}
		$bitmap[$Height -1 - $i] .= "\n";
	}
}

# ---------------------------------------------------------------------------
# 
sub write_palette
{
	my $tmp;
	print FILE "const u16 $constatntname\_pal[] = \n{\n";

	for($i = 0; $i < ($ColSize>>4); $i++) {
		$tmp = sprintf("\t/* %01X */\n\t", $i);
		print FILE "$tmp";
		for($j = 0; $j < 16; $j++) {
			$tmp = pop(@palette);
			$tmp = sprintf("%04x",$tmp);
			print FILE "0x$tmp,";
			print FILE "\n\t" if ($j == 7);
		}
		print FILE "\n";
	}
	print FILE "};\n\n";
}

# ---------------------------------------------------------------------------
# 
sub write_bitmap
{
	my $i, $j, $k, $l;
	my $tmp;
	my $number = 0;
	print FILE "const u16 $constatntname\_chr[] = \n{\n";
	for ($i = 0; $i < $Height; $i += 8) {

		for ($j = 0; $j < $Width; $j += 8) {
			print FILE "\t// $number\n";

			unless ($mode{'256'}) {
				# 16F[h
				for ($k = 0; $k < 8; $k++) {
					$tmp = substr($bitmap[$i + $k], $j,8);
					print FILE "\t" if ($k == 4 || $k == 0);

					for ($l = 0; $l < 8; $l += 4) {
						my $tmp2 = substr($tmp, $l+3,1);
						my $tmp3 = substr($tmp, $l+2,1);
						my $tmp4 = substr($tmp, $l+1,1);
						my $tmp5 = substr($tmp, $l  ,1);
						my $tmp6 = "0x".$tmp2.$tmp3.$tmp4.$tmp5.",";
						print FILE "$tmp6";
					}
					print FILE "\n" if ($k == 3);
				}
			} else {
				# 256F[h
				for ($k = 0; $k < 8; $k++) {
					$tmp = substr($bitmap[$i + $k], $j*2,16);
					print FILE "\t";
					for ($l = 0; $l < 16; $l += 4) {
						my $tmp2 = substr($tmp, $l  ,2);
						my $tmp3 = substr($tmp, $l+2,2);
						my $tmp4 = "0x".$tmp2.",0x".$tmp3.",";
						print FILE "$tmp4";
					}
					print FILE "\n";
				}
			}
			print FILE "\n";
			$number++;
		}
		print FILE "\n";

	}
	print FILE "};\n";
}

# ---------------------------------------------------------------------------
# 
sub prterror {
	print STDERR $_[0];
	exit 1;
}
